"""
UNETR based on: "Hatamizadeh et al.,
UNETR: Transformers for 3D Medical Image Segmentation <https://arxiv.org/abs/2103.10504>"
"""

from typing import Tuple
import torch
from monai.networks.blocks import UnetrBasicBlock, UnetrPrUpBlock, UnetrUpBlock
from monai.networks.blocks.dynunet_block import UnetOutBlock
from monai.networks.nets import ViT
from yucca.networks.networks.YuccaNet import YuccaNet


class UNetR(YuccaNet):
    def __init__(
        self,
        input_channels: int,
        patch_size: list | Tuple,
        dropout_op_kwargs={"p": 0.0},
        num_classes: int = 1,
        starting_filters: int = 16,
        weightInitializer=None,
    ) -> None:
        """
        Args:
            in_channels: dimension of input channels.
            num_classes: dimension of output channels.
            patch_size: dimension of input patch.
            starting_filters: dimension of network feature size.
            hidden_size: dimension of hidden layer.
            mlp_dim: dimension of feedforward layer.
            num_heads: number of attention heads.
            pos_embed: position embedding layer type.
            norm_name: feature normalization type and arguments.
            conv_block: bool argument to determine if convolutional block is used.
            res_block: bool argument to determine if residual block is used.

        Examples::

            # for single channel input 4-channel output with patch size of (96,96,96), feature size of 32 and batch norm
            >>> net = UNETR(in_channels=1, out_channels=4, img_size=(96,96,96), starting_filters=32, norm_name='batch')

            # for 4-channel input 3-channel output with patch size of (128,128,128), conv position embedding and instance norm
            >>> net = UNETR(in_channels=4, out_channels=3, img_size=(128,128,128), pos_embed='conv', norm_name='instance')

        """
        super(UNetR, self).__init__()
        # Fixed parameters
        hidden_size = 768
        mlp_dim = 3072
        num_heads = 12
        pos_embed = "conv"
        norm_name = "instance"
        conv_block = False
        res_block = True

        if not (0 <= dropout_op_kwargs["p"] <= 1):
            raise AssertionError("dropout_rate should be between 0 and 1.")

        if hidden_size % num_heads != 0:
            raise AssertionError("hidden size should be divisible by num_heads.")

        if pos_embed not in ["conv", "perceptron"]:
            raise KeyError(f"Position embedding layer of type {pos_embed} is not supported.")

        # Task specific
        self.num_classes = num_classes
        self.filters = starting_filters

        # Model parameters
        self.weightInitializer = weightInitializer

        self.num_layers = 12
        if len(patch_size) == 2:
            self.transformer_patch_size = (16, 16)
            self.feat_size = (
                patch_size[0] // self.transformer_patch_size[0],
                patch_size[1] // self.transformer_patch_size[1],
            )
            self.spatial_dims = 2
        else:
            self.transformer_patch_size = (16, 16, 16)
            self.feat_size = (
                patch_size[0] // self.transformer_patch_size[0],
                patch_size[1] // self.transformer_patch_size[1],
                patch_size[2] // self.transformer_patch_size[2],
            )
            self.spatial_dims = 3
        self.hidden_size = hidden_size
        self.classification = False

        self.vit = ViT(
            in_channels=input_channels,
            img_size=patch_size,
            patch_size=self.transformer_patch_size,
            hidden_size=hidden_size,
            mlp_dim=mlp_dim,
            num_layers=self.num_layers,
            num_heads=num_heads,
            pos_embed=pos_embed,
            classification=self.classification,
            dropout_rate=dropout_op_kwargs["p"],
            spatial_dims=self.spatial_dims,
        )
        self.encoder1 = UnetrBasicBlock(
            spatial_dims=self.spatial_dims,
            in_channels=input_channels,
            out_channels=self.filters,
            kernel_size=3,
            stride=1,
            norm_name=norm_name,
            res_block=res_block,
        )
        self.encoder2 = UnetrPrUpBlock(
            spatial_dims=self.spatial_dims,
            in_channels=hidden_size,
            out_channels=self.filters * 2,
            num_layer=2,
            kernel_size=3,
            stride=1,
            upsample_kernel_size=2,
            norm_name=norm_name,
            conv_block=conv_block,
            res_block=res_block,
        )
        self.encoder3 = UnetrPrUpBlock(
            spatial_dims=self.spatial_dims,
            in_channels=hidden_size,
            out_channels=self.filters * 4,
            num_layer=1,
            kernel_size=3,
            stride=1,
            upsample_kernel_size=2,
            norm_name=norm_name,
            conv_block=conv_block,
            res_block=res_block,
        )
        self.encoder4 = UnetrPrUpBlock(
            spatial_dims=self.spatial_dims,
            in_channels=hidden_size,
            out_channels=self.filters * 8,
            num_layer=0,
            kernel_size=3,
            stride=1,
            upsample_kernel_size=2,
            norm_name=norm_name,
            conv_block=conv_block,
            res_block=res_block,
        )
        self.decoder5 = UnetrUpBlock(
            spatial_dims=self.spatial_dims,
            in_channels=hidden_size,
            out_channels=self.filters * 8,
            kernel_size=3,
            upsample_kernel_size=2,
            norm_name=norm_name,
            res_block=res_block,
        )
        self.decoder4 = UnetrUpBlock(
            spatial_dims=self.spatial_dims,
            in_channels=self.filters * 8,
            out_channels=self.filters * 4,
            kernel_size=3,
            upsample_kernel_size=2,
            norm_name=norm_name,
            res_block=res_block,
        )
        self.decoder3 = UnetrUpBlock(
            spatial_dims=self.spatial_dims,
            in_channels=self.filters * 4,
            out_channels=self.filters * 2,
            kernel_size=3,
            upsample_kernel_size=2,
            norm_name=norm_name,
            res_block=res_block,
        )
        self.decoder2 = UnetrUpBlock(
            spatial_dims=self.spatial_dims,
            in_channels=self.filters * 2,
            out_channels=self.filters,
            kernel_size=3,
            upsample_kernel_size=2,
            norm_name=norm_name,
            res_block=res_block,
        )
        self.out = UnetOutBlock(spatial_dims=self.spatial_dims, in_channels=self.filters, out_channels=self.num_classes)  # type: ignore

        if self.weightInitializer is not None:
            print("initializing weights")
            self.apply(self.weightInitializer)

    def proj_feat(self, x, hidden_size, feat_size):
        if len(feat_size) == 2:
            x = x.view(x.size(0), feat_size[0], feat_size[1], hidden_size)
            x = x.permute(0, 3, 1, 2).contiguous()
        else:
            x = x.view(x.size(0), feat_size[0], feat_size[1], feat_size[2], hidden_size)
            x = x.permute(0, 4, 1, 2, 3).contiguous()
        return x

    def load_from(self, weights):
        with torch.no_grad():
            # copy weights from patch embedding
            for i in weights["state_dict"]:
                print(i)
            self.vit.patch_embedding.position_embeddings.copy_(
                weights["state_dict"]["module.transformer.patch_embedding.position_embeddings_3d"]
            )
            self.vit.patch_embedding.cls_token.copy_(weights["state_dict"]["module.transformer.patch_embedding.cls_token"])
            self.vit.patch_embedding.patch_embeddings[1].weight.copy_(
                weights["state_dict"]["module.transformer.patch_embedding.patch_embeddings.1.weight"]
            )
            self.vit.patch_embedding.patch_embeddings[1].bias.copy_(
                weights["state_dict"]["module.transformer.patch_embedding.patch_embeddings.1.bias"]
            )

            # copy weights from  encoding blocks (default: num of blocks: 12)
            for bname, block in self.vit.blocks.named_children():
                print(block)
                block.loadFrom(weights, n_block=bname)
            # last norm layer of transformer
            self.vit.norm.weight.copy_(weights["state_dict"]["module.transformer.norm.weight"])
            self.vit.norm.bias.copy_(weights["state_dict"]["module.transformer.norm.bias"])

    def forward(self, x_in):
        x, hidden_states_out = self.vit(x_in)
        enc1 = self.encoder1(x_in)
        x2 = hidden_states_out[3]
        enc2 = self.encoder2(self.proj_feat(x2, self.hidden_size, self.feat_size))
        x3 = hidden_states_out[6]
        enc3 = self.encoder3(self.proj_feat(x3, self.hidden_size, self.feat_size))
        x4 = hidden_states_out[9]
        enc4 = self.encoder4(self.proj_feat(x4, self.hidden_size, self.feat_size))
        dec4 = self.proj_feat(x, self.hidden_size, self.feat_size)
        dec3 = self.decoder5(dec4, enc4)
        dec2 = self.decoder4(dec3, enc3)
        dec1 = self.decoder3(dec2, enc2)
        out = self.decoder2(dec1, enc1)
        logits = self.out(out)
        return logits



from yucca.pipeline.managers.YuccaManager import YuccaManager
from yucca.optimization.loss_functions.deep_supervision import DeepSupervisionLoss


class YuccaManager_DS(YuccaManager):
    """
    The difference from YuccaTrainerV2 --> YuccaTrainerV3 is:
    - Introduces Deep Supervision
    - Uses data augmentation scheme V2
    """

    def __init__(
        self,
        model,
        model_dimensions: str,
        task: str,
        folds: str | int,
        plan_id: str,
        starting_lr: float = None,
        loss_fn: str = None,
        momentum: float = None,
        continue_training: bool = False,
        checkpoint: str = None,
        finetune: bool = False,
        fast_training: bool = False,
    ):
        super().__init__(
            model,
            model_dimensions,
            task,
            folds,
            plan_id,
            starting_lr,
            loss_fn,
            momentum,
            continue_training,
            checkpoint,
            finetune,
            fast_training,
        )
        self.deep_supervision = True

    def comprehensive_eval(self, pred, seg):
        pred = pred[0]
        seg = seg[0]
        super().comprehensive_eval(pred, seg)

    def initialize_loss_optim_lr(self):
        super().initialize_loss_optim_lr()
        self.loss_fn = DeepSupervisionLoss(self.loss_fn)

    def predict_folder(self, input_folder, output_folder, not_strict=True, save_softmax=False, overwrite=False, do_tta=True):
        ds = self.deep_supervision
        self.network.deep_supervision = False
        super().predict_folder(input_folder, output_folder, not_strict, save_softmax, overwrite, do_tta)
        self.network.deep_supervision = ds



import torch
import torch.nn as nn
from yucca.optimization.loss_functions.nnUNet_losses import get_tp_fp_fn_tn


class SoftSigmoidDiceLoss(nn.Module):
    def __init__(self, apply_nonlin=True, batch_dice=False, do_bg=True, smooth=1.0):
        """ """
        super(SoftSigmoidDiceLoss, self).__init__()

        self.do_bg = do_bg
        self.batch_dice = batch_dice
        self.smooth = smooth
        self.apply_nonlin = apply_nonlin

    def forward(self, x, y, loss_mask=None):
        shp_x = x.shape

        if self.batch_dice:
            axes = [0] + list(range(2, len(shp_x)))
        else:
            axes = list(range(2, len(shp_x)))

        if self.apply_nonlin is True:
            x = torch.sigmoid(x)

        tp, fp, fn, _ = get_tp_fp_fn_tn(x, y, axes, loss_mask, False)

        nominator = 2 * tp + self.smooth
        denominator = 2 * tp + fp + fn + self.smooth

        dc = nominator / (denominator + 1e-8)

        if not self.do_bg:
            if self.batch_dice:
                dc = dc[1:]
            else:
                dc = dc[:, 1:]
        dc = dc.mean()

        return -dc



"""
FROM: https://github.com/j-sripad/mulitresunet-pytorch/blob/main/multiresunet.py
"""

import torch
import torch.nn as nn
from yucca.networks.networks.YuccaNet import YuccaNet
from yucca.networks.blocks_and_layers.conv_blocks import (
    Multiresblock,
    Respath,
)
from yucca.networks.blocks_and_layers.conv_layers import (
    ConvDropoutNormNonlin,
)


class MultiResUNet(YuccaNet):
    """
    Arguments:
    channels - input image channels
    filters - filters to begin with (Unet)
    nclasses - number of classes

    Returns - None
    """

    def __init__(
        self,
        input_channels: int,
        num_classes: int = 1,
        starting_filters: int = 32,
        conv_op=nn.Conv2d,
        conv_kwargs={
            "kernel_size": 3,
            "stride": 1,
            "padding": 1,
            "dilation": 1,
            "bias": True,
        },
        norm_op=nn.InstanceNorm2d,
        norm_op_kwargs={"eps": 1e-5, "affine": True, "momentum": 0.1},
        dropout_op=nn.Dropout2d,
        dropout_op_kwargs={"p": 0.0, "inplace": True},
        nonlin=nn.LeakyReLU,
        nonlin_kwargs={"negative_slope": 1e-2, "inplace": True},
        dropout_in_decoder=False,
        weightInitializer=None,
        basic_block=ConvDropoutNormNonlin,
    ) -> None:
        super(MultiResUNet, self).__init__()
        # Architecture specific
        self.alpha = 1.67

        # Task specific
        self.num_classes = num_classes
        self.filters = starting_filters

        # Model parameters
        self.conv_op = conv_op
        self.conv_kwargs = conv_kwargs
        self.norm_op_kwargs = norm_op_kwargs
        self.norm_op = norm_op
        self.dropout_op = dropout_op
        self.dropout_op_kwargs = dropout_op_kwargs
        self.nonlin_kwargs = nonlin_kwargs
        self.nonlin = nonlin
        self.weightInitializer = weightInitializer
        self.basic_block = basic_block

        # Dimensions
        if self.conv_op == nn.Conv2d:
            self.pool_op = torch.nn.MaxPool2d
            self.upsample = torch.nn.ConvTranspose2d
        else:
            self.pool_op = torch.nn.MaxPool3d
            self.upsample = torch.nn.ConvTranspose3d

        # Encoder Path
        self.multiresblock1 = Multiresblock(
            input_channels,
            self.filters,
            self.conv_op,
            self.conv_kwargs,
            self.norm_op,
            self.norm_op_kwargs,
            self.dropout_op,
            self.dropout_op_kwargs,
            self.nonlin,
            self.nonlin_kwargs,
            self.basic_block,
        )

        self.in_filters1 = (
            int(self.filters * self.alpha * 0.167)
            + int(self.filters * self.alpha * 0.333)
            + int(self.filters * self.alpha * 0.5)
        )
        self.pool1 = self.pool_op(2)
        self.respath1 = Respath(
            self.in_filters1,
            self.filters,
            4,
            self.conv_op,
            self.conv_kwargs,
            self.norm_op,
            self.norm_op_kwargs,
            self.dropout_op,
            self.dropout_op_kwargs,
            self.nonlin,
            self.nonlin_kwargs,
            self.basic_block,
        )

        self.multiresblock2 = Multiresblock(
            self.in_filters1,
            self.filters * 2,
            self.conv_op,
            self.conv_kwargs,
            self.norm_op,
            self.norm_op_kwargs,
            self.dropout_op,
            self.dropout_op_kwargs,
            self.nonlin,
            self.nonlin_kwargs,
            self.basic_block,
        )

        self.in_filters2 = (
            int(self.filters * 2 * self.alpha * 0.167)
            + int(self.filters * 2 * self.alpha * 0.333)
            + int(self.filters * 2 * self.alpha * 0.5)
        )
        self.pool2 = self.pool_op(2)
        self.respath2 = Respath(
            self.in_filters2,
            self.filters * 2,
            3,
            self.conv_op,
            self.conv_kwargs,
            self.norm_op,
            self.norm_op_kwargs,
            self.dropout_op,
            self.dropout_op_kwargs,
            self.nonlin,
            self.nonlin_kwargs,
            self.basic_block,
        )

        self.multiresblock3 = Multiresblock(
            self.in_filters2,
            self.filters * 4,
            self.conv_op,
            self.conv_kwargs,
            self.norm_op,
            self.norm_op_kwargs,
            self.dropout_op,
            self.dropout_op_kwargs,
            self.nonlin,
            self.nonlin_kwargs,
            self.basic_block,
        )

        self.in_filters3 = (
            int(self.filters * 4 * self.alpha * 0.167)
            + int(self.filters * 4 * self.alpha * 0.333)
            + int(self.filters * 4 * self.alpha * 0.5)
        )
        self.pool3 = self.pool_op(2)
        self.respath3 = Respath(
            self.in_filters3,
            self.filters * 4,
            2,
            self.conv_op,
            self.conv_kwargs,
            self.norm_op,
            self.norm_op_kwargs,
            self.dropout_op,
            self.dropout_op_kwargs,
            self.nonlin,
            self.nonlin_kwargs,
            self.basic_block,
        )

        self.multiresblock4 = Multiresblock(
            self.in_filters3,
            self.filters * 8,
            self.conv_op,
            self.conv_kwargs,
            self.norm_op,
            self.norm_op_kwargs,
            self.dropout_op,
            self.dropout_op_kwargs,
            self.nonlin,
            self.nonlin_kwargs,
            self.basic_block,
        )

        self.in_filters4 = (
            int(self.filters * 8 * self.alpha * 0.167)
            + int(self.filters * 8 * self.alpha * 0.333)
            + int(self.filters * 8 * self.alpha * 0.5)
        )
        self.pool4 = self.pool_op(2)
        self.respath4 = Respath(
            self.in_filters4,
            self.filters * 8,
            1,
            self.conv_op,
            self.conv_kwargs,
            self.norm_op,
            self.norm_op_kwargs,
            self.dropout_op,
            self.dropout_op_kwargs,
            self.nonlin,
            self.nonlin_kwargs,
            self.basic_block,
        )

        self.multiresblock5 = Multiresblock(
            self.in_filters4,
            self.filters * 16,
            self.conv_op,
            self.conv_kwargs,
            self.norm_op,
            self.norm_op_kwargs,
            self.dropout_op,
            self.dropout_op_kwargs,
            self.nonlin,
            self.nonlin_kwargs,
            self.basic_block,
        )

        if not dropout_in_decoder:
            old_dropout_p = self.dropout_op_kwargs["p"]
            self.dropout_op_kwargs["p"] = 0.0

        # Decoder path
        self.in_filters5 = (
            int(self.filters * 16 * self.alpha * 0.167)
            + int(self.filters * 16 * self.alpha * 0.333)
            + int(self.filters * 16 * self.alpha * 0.5)
        )
        self.upsample6 = self.upsample(self.in_filters5, self.filters * 8, kernel_size=(2), stride=(2))

        self.multiresblock6 = Multiresblock(
            self.filters * 8 * 2,
            self.filters * 8,
            self.conv_op,
            self.conv_kwargs,
            self.norm_op,
            self.norm_op_kwargs,
            self.dropout_op,
            self.dropout_op_kwargs,
            self.nonlin,
            self.nonlin_kwargs,
            self.basic_block,
        )

        self.in_filters6 = (
            int(self.filters * 8 * self.alpha * 0.167)
            + int(self.filters * 8 * self.alpha * 0.333)
            + int(self.filters * 8 * self.alpha * 0.5)
        )
        self.upsample7 = self.upsample(self.in_filters6, self.filters * 4, kernel_size=(2), stride=(2))

        self.multiresblock7 = Multiresblock(
            self.filters * 4 * 2,
            self.filters * 4,
            self.conv_op,
            self.conv_kwargs,
            self.norm_op,
            self.norm_op_kwargs,
            self.dropout_op,
            self.dropout_op_kwargs,
            self.nonlin,
            self.nonlin_kwargs,
            self.basic_block,
        )

        self.in_filters7 = (
            int(self.filters * 4 * self.alpha * 0.167)
            + int(self.filters * 4 * self.alpha * 0.333)
            + int(self.filters * 4 * self.alpha * 0.5)
        )
        self.upsample8 = self.upsample(self.in_filters7, self.filters * 2, kernel_size=(2), stride=(2))

        self.multiresblock8 = Multiresblock(
            self.filters * 2 * 2,
            self.filters * 2,
            self.conv_op,
            self.conv_kwargs,
            self.norm_op,
            self.norm_op_kwargs,
            self.dropout_op,
            self.dropout_op_kwargs,
            self.nonlin,
            self.nonlin_kwargs,
            self.basic_block,
        )

        self.in_filters8 = (
            int(self.filters * 2 * self.alpha * 0.167)
            + int(self.filters * 2 * self.alpha * 0.333)
            + int(self.filters * 2 * self.alpha * 0.5)
        )
        self.upsample9 = self.upsample(self.in_filters8, self.filters, kernel_size=(2), stride=(2))

        self.multiresblock9 = Multiresblock(
            self.filters * 2,
            self.filters,
            self.conv_op,
            self.conv_kwargs,
            self.norm_op,
            self.norm_op_kwargs,
            self.dropout_op,
            self.dropout_op_kwargs,
            self.nonlin,
            self.nonlin_kwargs,
            self.basic_block,
        )

        self.in_filters9 = (
            int(self.filters * self.alpha * 0.167)
            + int(self.filters * self.alpha * 0.333)
            + int(self.filters * self.alpha * 0.5)
        )
        self.conv_final = self.conv_op(self.in_filters9, self.num_classes, 1, 1, 0, 1, 1, False)

        if not dropout_in_decoder:
            self.dropout_op_kwargs["p"] = old_dropout_p

        if self.weightInitializer is not None:
            print("initializing weights")
            self.apply(self.weightInitializer)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x_multires1 = self.multiresblock1(x)
        x_pool1 = self.pool1(x_multires1)
        x_multires1 = self.respath1(x_multires1)

        x_multires2 = self.multiresblock2(x_pool1)
        x_pool2 = self.pool2(x_multires2)
        x_multires2 = self.respath2(x_multires2)

        x_multires3 = self.multiresblock3(x_pool2)
        x_pool3 = self.pool3(x_multires3)
        x_multires3 = self.respath3(x_multires3)

        x_multires4 = self.multiresblock4(x_pool3)
        x_pool4 = self.pool4(x_multires4)
        x_multires4 = self.respath4(x_multires4)

        x_multires5 = self.multiresblock5(x_pool4)

        up6 = torch.cat([self.upsample6(x_multires5), x_multires4], axis=1)
        x_multires6 = self.multiresblock6(up6)
        up7 = torch.cat([self.upsample7(x_multires6), x_multires3], axis=1)
        x_multires7 = self.multiresblock7(up7)

        up8 = torch.cat([self.upsample8(x_multires7), x_multires2], axis=1)
        x_multires8 = self.multiresblock8(up8)

        up9 = torch.cat([self.upsample9(x_multires8), x_multires1], axis=1)
        x_multires9 = self.multiresblock9(up9)

        logits = self.conv_final(x_multires9)

        return logits



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Dec 16 13:25:09 2022

@author: zcr545
"""
from yucca.functional.array_operations.bounding_boxes import get_bbox_for_label


class Box(object):
    def __init__(self, array, label, padding):
        # Create bounding box for the object.
        self.array = array
        self.label = label
        self.padding = padding
        self.set_coordinates()

    def set_coordinates(self):
        box = get_bbox_for_label(self.array, self.label, self.padding)

        assert len(box) in [4, 6], "invalid box dimensions. Should be " f"4 (2D) or 6 (3D) but is {len(box)}"
        if len(box) == 6:
            self.xmin, self.xmax, self.ymin, self.ymax, self.zmin, self.zmax = box
        if len(box) == 4:
            self.xmin, self.xmax, self.ymin, self.ymax = box

    @property
    def area(self):
        """
        Calculates the surface area. useful for IOU!
        """
        width = abs(self.xmax - self.xmin)
        height = abs(self.ymax - self.ymin)
        depth = abs(self.zmax - self.zmin)
        return width * height * depth

    def intersect(self, bbox):
        # Checks that exclude intersection
        if self.xmin > bbox.xmax or bbox.xmin > self.xmax:
            return False

        if self.ymin > bbox.ymax or bbox.ymin > self.ymax:
            return False

        if self.zmin > bbox.zmax or bbox.zmin > self.zmax:
            return False

        x1 = min(self.xmax, bbox.xmax)
        x2 = max(self.xmin, bbox.xmin)

        y1 = min(self.ymax, bbox.ymax)
        y2 = max(self.ymin, bbox.ymin)

        z1 = min(self.zmax, bbox.zmax)
        z2 = max(self.zmin, bbox.zmin)

        intersection = max(x1 - x2, 0) * max(y1 - y2, 0) * max(z1 - z2, 0)
        return intersection

    def iou(self, bbox):
        intersection = self.intersect(bbox)
        if not intersection:
            return False

        iou = intersection / float(self.area + bbox.area - intersection)
        # return the intersection over union value
        return iou



from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p, subdirs
from yucca.pipeline.task_conversion.utils import generate_dataset_json
from yucca.paths import yucca_raw_data
import shutil


def convert(path: str, subdir: str = "MSSEG1_2016"):
    # Target names
    task_name = "Task011_MSSEG1"
    task_prefix = "MSSEG1"

    ###OUTPUT DATA
    # Target paths
    target_base = join(yucca_raw_data, task_name)

    target_imagesTr = join(target_base, "imagesTr")
    target_labelsTr = join(target_base, "labelsTr")

    target_imagesTs = join(target_base, "imagesTs")
    target_labelsTs = join(target_base, "labelsTs")

    maybe_mkdir_p(target_imagesTr)
    maybe_mkdir_p(target_labelsTs)
    maybe_mkdir_p(target_imagesTs)
    maybe_mkdir_p(target_labelsTr)

    # MSSEG 2016
    # INPUT DATA
    # Input path and names
    base_in = join(path, subdir)
    file_suffix = ".nii.gz"

    # Train/Test Splits
    # We only use the train folder, as the test folder does not contain segmentations
    # to obtain those, submission to the challenge is required (it's from 2008, forget it)
    train_folder = join(base_in, "training")
    test_folder = join(base_in, "Testing")

    ###Populate Target Directory###
    for center in subdirs(train_folder, join=False):
        training_samples = subdirs(join(train_folder, center), join=False)
        for sTr in training_samples:
            src_image_file_path1 = join(train_folder, center, sTr, "Preprocessed_Data", "DP_preprocessed" + file_suffix)
            src_image_file_path2 = join(train_folder, center, sTr, "Preprocessed_Data", "FLAIR_preprocessed" + file_suffix)
            src_image_file_path3 = join(train_folder, center, sTr, "Preprocessed_Data", "GADO_preprocessed" + file_suffix)
            src_image_file_path4 = join(train_folder, center, sTr, "Preprocessed_Data", "T1_preprocessed" + file_suffix)
            src_image_file_path5 = join(train_folder, center, sTr, "Preprocessed_Data", "T2_preprocessed" + file_suffix)
            dst_image_file_path1 = f"{target_imagesTr}/{task_prefix}_{center}_{sTr}_000.nii.gz"
            dst_image_file_path2 = f"{target_imagesTr}/{task_prefix}_{center}_{sTr}_001.nii.gz"
            dst_image_file_path3 = f"{target_imagesTr}/{task_prefix}_{center}_{sTr}_002.nii.gz"
            dst_image_file_path4 = f"{target_imagesTr}/{task_prefix}_{center}_{sTr}_003.nii.gz"
            dst_image_file_path5 = f"{target_imagesTr}/{task_prefix}_{center}_{sTr}_004.nii.gz"

            src_label_path = join(train_folder, center, sTr, "Masks", "Consensus" + file_suffix)
            dst_label_path = f"{target_labelsTr}/{task_prefix}_{center}_{sTr}.nii.gz"

            shutil.copy2(src_image_file_path1, dst_image_file_path1)
            shutil.copy2(src_image_file_path2, dst_image_file_path2)
            shutil.copy2(src_image_file_path3, dst_image_file_path3)
            shutil.copy2(src_image_file_path4, dst_image_file_path4)
            shutil.copy2(src_image_file_path5, dst_image_file_path5)

            shutil.copy2(src_label_path, dst_label_path)

    for center in subdirs(test_folder, join=False):
        test_samples = subdirs(join(test_folder, center), join=False)
        for sTs in test_samples:
            src_image_file_path1 = join(test_folder, center, sTs, "Preprocessed_Data", "DP_preprocessed" + file_suffix)
            src_image_file_path2 = join(test_folder, center, sTs, "Preprocessed_Data", "FLAIR_preprocessed" + file_suffix)
            src_image_file_path3 = join(test_folder, center, sTs, "Preprocessed_Data", "GADO_preprocessed" + file_suffix)
            src_image_file_path4 = join(test_folder, center, sTs, "Preprocessed_Data", "T1_preprocessed" + file_suffix)
            src_image_file_path5 = join(test_folder, center, sTs, "Preprocessed_Data", "T2_preprocessed" + file_suffix)
            dst_image_file_path1 = f"{target_imagesTs}/{task_prefix}_{center}_{sTs}_000.nii.gz"
            dst_image_file_path2 = f"{target_imagesTs}/{task_prefix}_{center}_{sTs}_001.nii.gz"
            dst_image_file_path3 = f"{target_imagesTs}/{task_prefix}_{center}_{sTs}_002.nii.gz"
            dst_image_file_path4 = f"{target_imagesTs}/{task_prefix}_{center}_{sTs}_003.nii.gz"
            dst_image_file_path5 = f"{target_imagesTs}/{task_prefix}_{center}_{sTs}_004.nii.gz"

            src_label_path = join(test_folder, center, sTs, "Masks", "Consensus" + file_suffix)
            dst_label_path = f"{target_labelsTs}/{task_prefix}_{center}_{sTs}.nii.gz"

            shutil.copy2(src_image_file_path1, dst_image_file_path1)
            shutil.copy2(src_image_file_path2, dst_image_file_path2)
            shutil.copy2(src_image_file_path3, dst_image_file_path3)
            shutil.copy2(src_image_file_path4, dst_image_file_path4)
            shutil.copy2(src_image_file_path5, dst_image_file_path5)

            shutil.copy2(src_label_path, dst_label_path)

    generate_dataset_json(
        join(target_base, "dataset.json"),
        target_imagesTr,
        target_imagesTs,
        ("DP", "FLAIR", "GADO", "T1", "T2"),
        labels={0: "background", 1: "MS Lesion"},
        dataset_name=task_name,
        license="hands off!",
        dataset_description="MSSEG1",
        dataset_reference="https://www.nitrc.org/projects/msseg, https://arxiv.org/abs/2206.06694",
    )


if __name__ == "__main__":
    convert()



from yucca.data.augmentation.transforms.YuccaTransform import YuccaTransform
from yucca.functional.transforms import simulate_lowres
import numpy as np


class SimulateLowres(YuccaTransform):
    def __init__(
        self,
        data_key="image",
        p_per_sample: float = 1.0,
        p_per_channel: float = 0.5,
        p_per_axis: float = 0.33,
        zoom_range=(0.5, 1.0),
        clip_to_input_range=False,
    ):
        self.data_key = data_key
        self.p_per_sample = p_per_sample
        self.p_per_channel = p_per_channel
        self.p_per_axis = p_per_axis
        self.zoom_range = zoom_range
        self.clip_to_input_range = clip_to_input_range

    @staticmethod
    def get_params(zoom_range, shape, p_per_axis):
        # No parameters to retrieve
        if isinstance(shape, (list, tuple)):
            shape = np.array(shape)
        zoom = np.random.uniform(*zoom_range)
        dim = len(shape)
        zoomed_shape = np.round(shape * zoom).astype(int)
        for i in range(dim):
            if np.random.uniform() < p_per_axis:
                shape[i] = zoomed_shape[i]
        return shape

    def __simulatelowres__(self, image, target_shape):
        image = simulate_lowres(image, target_shape=target_shape, clip_to_input_range=self.clip_to_input_range)
        return image

    def __call__(self, packed_data_dict=None, **unpacked_data_dict):
        data_dict = packed_data_dict if packed_data_dict else unpacked_data_dict
        assert (
            len(data_dict[self.data_key].shape) == 5 or len(data_dict[self.data_key].shape) == 4
        ), f"Incorrect data size or shape.\
            \nShould be (b, c, x, y, z) or (b, c, x, y) and is: {data_dict[self.data_key].shape}"

        for b in range(data_dict[self.data_key].shape[0]):
            if np.random.uniform() < self.p_per_sample:
                for c in range(data_dict[self.data_key][b].shape[0]):
                    if np.random.uniform() < self.p_per_channel:
                        target_shape = self.get_params(
                            self.zoom_range,
                            data_dict[self.data_key][b, c].shape,
                            self.p_per_axis,
                        )
                        data_dict[self.data_key][b, c] = self.__simulatelowres__(data_dict[self.data_key][b, c], target_shape)
        return data_dict



import warnings
from skimage import exposure
import numpy as np
from typing import Optional


def normalizer(array: np.ndarray, scheme: str, intensities: Optional[dict] = None):
    """
    Normalizing function for preprocessing and inference.

    supported schemes can be either:
    None = for no normalization. Generally not recommended.
    MinMax = for 0-1 or Min-Max normalization.
    255to1 = for 0-1 normalization of 8 bit images.
    Standardize = (array - mean) / std. Based on modality wide stats.
    Clip = for contrast clipping. This will clip values to the 0.01 and 99.99th percentiles
        and then perform 0-1 normalization.
    """
    accepted_schemes = ["clipping", "minmax", "range", "no_norm", "standardize", "volume_wise_znorm", "255to1"]

    assert scheme in accepted_schemes, "invalid normalization scheme inserted" f"attempted scheme: {scheme}"
    assert array is not None

    if scheme == "no_norm":
        return array

    elif scheme == "minmax":
        assert intensities is not None, "ERROR: dataset wide stats are required for minmax"
        return (array - intensities["min"]) / (intensities["max"] - intensities["min"] + 1e-9)

    elif scheme == "255to1":
        return array / 255

    elif scheme == "range":
        print(array.max(), array.min(), intensities)
        return (array - array.min()) / (array.max() - array.min() + 1e-9) * (
            intensities["max"] - intensities["min"] + 1e-9
        ) + intensities["min"]

    elif scheme == "standardize":
        assert intensities is not None, "ERROR: dataset wide stats are required for standardize"
        return (array - float(intensities["mean"])) / float(intensities["std"])

    elif scheme == "clip":
        lower_bound, upper_bound = np.percentile(array, (0.01, 99.99))
        array = exposure.rescale_intensity(array, in_range=(lower_bound, upper_bound), out_range=(0, 1))
        return array

    elif scheme == "volume_wise_znorm":
        empty_val = array.min()  # We assume the background is the minimum value
        if empty_val != array[0, 0, 0]:
            warnings.warn(
                "Tried to normalize an array where the top right value was not the same as the minimum value."
                f"empty_val: {empty_val}, top right: {array[0, 0, 0]}"
            )
        mask = array != empty_val
        array = clamp(array, mask=mask)
        array = znormalize(array, mask=mask)
        array = rescale(array, range=(0, 1))
        return array


def clamp(x, mask, q=0.99):
    q_val = np.quantile(x[mask], q)
    return np.clip(x, a_min=None, a_max=q_val)


def znormalize(x, mask):
    values = x[mask]
    mean, std = np.mean(values), np.std(values)
    assert std > 0
    x -= mean
    x /= std
    return x


def rescale(x, range=(0, 1)):
    return exposure.rescale_intensity(x, out_range=range)



from batchgenerators.utilities.file_and_folder_operations import load_json


class node:
    def __init__(self, name):
        self.name = name
        self.children = []
        self.channel = None
        self.layer = None


class LossTree:
    def __init__(self, TreeDict):
        if isinstance(TreeDict, str):
            TreeDict = load_json(TreeDict)
        self.root = node("Tree Root")
        self.layers = []
        self.loss_layers = []
        self.recursive_expand(TreeDict)
        self.layers = set(self.layers)
        self.get_loss_layers()

    def recursive_expand(self, treedict, starting_node=None, starting_layer=-1):
        starting_layer += 1
        if not starting_node:
            starting_node = self.root
        for key, value in treedict.items():
            newnode = node(key)
            newnode.layer = starting_layer
            self.layers.append(newnode.layer)
            starting_node.children.append(newnode)
            if not isinstance(value, dict):
                newnode.channel = value
            else:
                self.recursive_expand(value, starting_node=newnode, starting_layer=starting_layer)

    def get_loss_layers(self):
        for layer in self.layers:
            layer_list = []
            self.loss_layers.append(self.recursive_loss_layers(layer, self.root, layer_list))

    def get_subclasses(self, node, class_list):
        if len(node.children) == 0:
            class_list.append(node.channel)
        else:
            for child in node.children:
                self.get_subclasses(child, class_list)
        return class_list

    def print(self, node=None, indent=0):
        if not node:
            node = self.root
            prefix = ""
        else:
            prefix = "|- " + str(node.layer) + ". "

        if len(node.children) == 0:
            print(" " * indent + prefix + node.name + " = " + str(node.channel))
        else:
            print(" " * indent + prefix + node.name)
            for child in node.children:
                self.print(child, indent + 4)

    def recursive_loss_layers(self, layer, starting_node, layer_list):
        for child in starting_node.children:
            if len(child.children) == 0 or child.layer == layer:
                layer_list.append(self.get_subclasses(child, []))
            else:
                self.recursive_loss_layers(layer, child, layer_list=layer_list)
        return layer_list



from .planning import make_plans_file, add_stats_to_plans_post_preprocessing
from .preprocessing import (
    analyze_label,
    determine_resample_size_from_target_size,
    determine_resample_size_from_target_spacing,
    determine_target_size,
    resample_and_normalize_case,
    pad_case_to_size,
    apply_nifti_preprocessing_and_return_numpy,
    preprocess_case_for_inference,
    preprocess_case_for_training_with_label,
    preprocess_case_for_training_without_label,
    reverse_preprocessing,
)



from yucca.data.augmentation.transforms.YuccaTransform import YuccaTransform
from yucca.functional.array_operations.normalization import normalizer


class Normalize(YuccaTransform):
    def __init__(
        self, normalize: bool = False, data_key: str = "image", metadata_key="metadata", scheme: str = "volume_wise_znorm"
    ):
        assert scheme in [
            "255to1",
            "minmax",
            "range",
            "clip",
            "volume_wise_znorm",
        ], f"Scheme {scheme} not supported, only schemes which are not relying on dataset level properties are currently supported"

        self.normalize = normalize
        self.data_key = data_key
        self.metadata_key = metadata_key
        self.scheme = scheme

    @staticmethod
    def get_params():
        # No parameters to retrieve
        pass

    def __normalize__(self, data_dict):
        for b in range(data_dict[self.data_key].shape[0]):
            for c in range(data_dict[self.data_key].shape[1]):
                data_dict[self.data_key][b, c] = normalizer(
                    data_dict[self.data_key][b, c], scheme=self.scheme, intensities=data_dict[self.metadata_key]
                )
        return data_dict

    def __call__(self, packed_data_dict=None, **unpacked_data_dict):
        data_dict = packed_data_dict if packed_data_dict else unpacked_data_dict
        assert (
            len(data_dict[self.data_key].shape) == 5 or len(data_dict[self.data_key].shape) == 4
        ), f"Incorrect data size or shape.\
            \nShould be (b, c, x, y, z) or (b, c, x, y) and is: {data_dict[self.data_key].shape}"

        if self.normalize:
            data_dict = self.__normalize__(data_dict)
        return data_dict



import sys
import os
import logging
from argparse import Namespace
from lightning.pytorch.loggers.logger import Logger
from pytorch_lightning.utilities.rank_zero import rank_zero_only
from pytorch_lightning.core.saving import save_hparams_to_yaml
from lightning_fabric.utilities.logger import _convert_params
from time import localtime, strftime, time
from batchgenerators.utilities.file_and_folder_operations import (
    join,
    maybe_mkdir_p,
    isdir,
)
from typing import Any, Dict, Optional, Union


class YuccaLogger(Logger):
    def __init__(
        self,
        disable_logging: bool = False,
        save_dir: str = "./",
        name: str = "",
        steps_per_epoch: int = None,
        version: Optional[Union[int, str]] = None,
    ):
        super().__init__()
        self._name = name or ""
        self._root_dir = save_dir
        self._version = version
        self.disable_logging = disable_logging
        self.steps_per_epoch = steps_per_epoch

        # Default params
        self.epoch_start_time = time()
        self.log_file = None
        self.previous_epoch = 0
        self.hparams: Dict[str, Any] = {}
        self.NAME_HPARAMS_FILE = "hparams.yaml"

        if self.disable_logging is False:
            if self.log_file is None:
                self.create_logfile()
            self.duplicate_console_out_to_log_file(self.log_file)

    @property
    def name(self):
        return self._name

    @property
    def version(self):
        return self._version

    @property
    def root_dir(self):
        return self._root_dir

    @property
    def log_dir(self):
        log_dir = self.root_dir
        if self.name is not None:
            log_dir = join(log_dir, self.name)
        if self.version is not None:
            version = self.version if isinstance(self.version, str) else f"version_{self.version}"
            log_dir = join(log_dir, version)
        if not isdir(log_dir):
            maybe_mkdir_p(log_dir)
        return log_dir

    @rank_zero_only
    def create_logfile(self):
        maybe_mkdir_p(self.log_dir)
        self.log_file = join(
            self.log_dir,
            "training_log.txt",
        )
        with open(self.log_file, "w") as f:
            f.write("Starting model training")
            logging.info("Starting model training \n" f'{"log file:":20} {self.log_file} \n')
            f.write("\n")
            f.write(f'{"log file:":20} {self.log_file}')
            f.write("\n")

    @rank_zero_only
    def duplicate_console_out_to_log_file(self, log_file):
        # Add the log_file as a duplicate handler of lightning outputs
        logging.getLogger("lightning.pytorch").addHandler(logging.FileHandler(log_file))
        # Add the log_file as a duplicate handler of lightning outputs
        logging.getLogger().addHandler(logging.FileHandler(log_file))

    @rank_zero_only
    def log_hyperparams(self, params: Union[Dict[str, Any], Namespace]) -> None:  # type: ignore[override]
        params = _convert_params(params)
        self.log_hparams(params)

    @rank_zero_only
    def log_metrics(self, metrics, step):
        if self.disable_logging:
            return
        # metrics is a dictionary of metric names and values
        # your code to record metrics goes here
        t = strftime("%Y_%m_%d_%H_%M_%S", localtime())
        with open(self.log_file, "a+") as f:
            current_epoch = (step + 1) // self.steps_per_epoch
            if current_epoch != self.previous_epoch:
                epoch_end_time = time()
                f.write("\n")
                f.write("\n")
                print("\n")
                f.write(f"{t} {'Current Epoch:':20} {current_epoch} \n")
                f.write(f"{t} {'Epoch Time:':20} {epoch_end_time-self.epoch_start_time} \n")
                print(f"{t} {'Current Epoch:':20} {current_epoch}")
                print(f"{t} {'Epoch Time:':20} {epoch_end_time-self.epoch_start_time}")
                self.previous_epoch = current_epoch
                self.epoch_start_time = epoch_end_time
            for key in metrics:
                if key == "epoch":
                    continue
                f.write(f"{t} {key+':':20} {metrics[key]} \n")
                print(f"{t} {key+':':20} {metrics[key]}")
        sys.stdout.flush()

    @rank_zero_only
    def log_hparams(self, params: Dict[str, Any]) -> None:
        """Record hparams."""
        self.hparams.update(params)

    @rank_zero_only
    def save(self) -> None:
        """Save recorded hparams into yaml."""
        hparams_file = os.path.join(self.log_dir, self.NAME_HPARAMS_FILE)
        save_hparams_to_yaml(hparams_file, self.hparams)

    @rank_zero_only
    def finalize(self, _status) -> None:
        # When using multiprocessing, finalize() should be a no-op on the main process, as no experiment has been
        # initialized there
        self.save()



import nibabel as nib
import shutil
from sklearn.model_selection import train_test_split
from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p, subfiles
from yucca.pipeline.task_conversion.utils import generate_dataset_json
from yucca.paths import yucca_raw_data
from yucca.functional.testing.data.nifti import verify_spacing_is_equal, verify_orientation_is_equal

# INPUT DATA
# Define input path and extension

folder_with_images = "/home/zcr545/datasets/decathlon/Task02_Heart"
file_extension = ".nii.gz"

# OUTPUT DATA
# Define the task name and prefix
task_name = "Task022_Heart"
task_prefix = "Heart"

# Set target paths
target_base = join(yucca_raw_data, task_name)
target_imagesTr = join(target_base, "imagesTr")
target_labelsTr = join(target_base, "labelsTr")
target_imagesTs = join(target_base, "imagesTs")
target_labelsTs = join(target_base, "labelsTs")

maybe_mkdir_p(target_imagesTr)
maybe_mkdir_p(target_labelsTs)
maybe_mkdir_p(target_imagesTs)
maybe_mkdir_p(target_labelsTr)

# Split data
images_dir = join(folder_with_images, "imagesTr")
labels_dir = join(folder_with_images, "labelsTr")
samples = subfiles(labels_dir, join=False, suffix=file_extension)
train_samples, test_samples = train_test_split(samples, test_size=0.2, random_state=1243)
images_dir_tr = images_dir_ts = images_dir
labels_dir_tr = labels_dir_ts = labels_dir


# Populate Target Directory
# This is also the place to apply any re-orientation, resampling and/or label correction.

for sTr in train_samples:
    image_path = join(images_dir_tr, sTr)
    label_path = join(labels_dir_tr, sTr)
    sTr = sTr[: -len(file_extension)]

    image = nib.load(image_path)
    label = nib.load(label_path)
    verify_spacing_is_equal(image, label)
    verify_orientation_is_equal(image, label)

    shutil.copy2(image_path, filename=f"{target_imagesTr}/{task_prefix}_{sTr}_000.nii.gz")
    shutil.copy2(label_path, filename=f"{target_labelsTr}/{task_prefix}_{sTr}.nii.gz")

for sTs in test_samples:
    image_path = join(images_dir_ts, sTs)
    label_path = join(labels_dir_ts, sTs)
    sTs = sTs[: -len(file_extension)]

    image = nib.load(image_path)
    label = nib.load(label_path)
    verify_spacing_is_equal(image, label)
    verify_orientation_is_equal(image, label)

    shutil.copy2(image_path, filename=f"{target_imagesTs}/{task_prefix}_{sTs}_000.nii.gz")
    shutil.copy2(label_path, filename=f"{target_labelsTs}/{task_prefix}_{sTs}.nii.gz")

generate_dataset_json(
    join(target_base, "dataset.json"),
    target_imagesTr,
    target_imagesTs,
    modalities=("T1",),
    labels={0: "Background", 1: "Left Atrium"},
    dataset_name=task_name,
    license="CC-BY-SA 4.0",
    dataset_description="Decathlon: Left Atrium Segmentation",
    dataset_reference="King's College London",
)



no_aug = {
    "additive_noise_p_per_sample": 0.0,
    "biasfield_p_per_sample": 0.0,
    "blurring_p_per_sample": 0.0,
    "blurring_p_per_channel": 0.0,
    "elastic_deform_p_per_sample": 0.0,
    "gamma_p_per_sample": 0.0,
    "gamma_p_invert_image": 0.0,
    "gibbs_ringing_p_per_sample": 0.0,
    "mirror_p_per_sample": 0.0,
    "mirror_p_per_axis": 0.0,
    "motion_ghosting_p_per_sample": 0.0,
    "multiplicative_noise_p_per_sample": 0.0,
    "rotation_p_per_sample": 0.0,
    "rotation_p_per_axis": 0.0,
    "scale_p_per_sample": 0.0,
    "simulate_lowres_p_per_sample": 0.0,
    "simulate_lowres_p_per_channel": 0.0,
    "simulate_lowres_p_per_axis": 0.0,
}

all_always = {
    "random_crop": True,
    "mask_image_for_reconstruction": True,
    "skip_label": False,
    "label_dtype": int,
    "copy_image_to_label": False,
    "additive_noise_p_per_sample": 1.0,
    "biasfield_p_per_sample": 1.0,
    "blurring_p_per_sample": 1.0,
    "blurring_p_per_channel": 1.0,
    "elastic_deform_p_per_sample": 1.0,
    "gamma_p_per_sample": 1.0,
    "gamma_p_invert_image": 1.0,
    "gibbs_ringing_p_per_sample": 1.0,
    "mirror_p_per_sample": 1.0,
    "mirror_p_per_axis": 1.0,
    "motion_ghosting_p_per_sample": 1.0,
    "multiplicative_noise_p_per_sample": 1.0,
    "rotation_p_per_sample": 1.0,
    "rotation_p_per_axis": 1.0,
    "scale_p_per_sample": 1.0,
    "simulate_lowres_p_per_sample": 1.0,
    "simulate_lowres_p_per_channel": 1.0,
    "simulate_lowres_p_per_axis": 1.0,
}


basic = no_aug.copy()
basic["mirror_p_per_sample"] = 0.2
basic["mirror_p_per_axis"] = 0.66
basic["rotation_p_per_sample"] = 0.2
basic["rotation_p_per_axis"] = 0.66
basic["scale_p_per_sample"] = 0.2


CT = {}
CT["gibbs_ringing_p_per_sample"] = 0.0
CT["motion_ghosting_p_per_sample"] = 0.0
CT["biasfield_p_per_sample"] = 0.0

generic = {
    "random_crop": False,
    "mask_image_for_reconstruction": False,
    "clip_to_input_range": False,  # ensures no augmentations go beyond the input range of the image/patch
    "normalize": False,
    # label/segmentation transforms
    "skip_label": False,
    "label_dtype": int,
    "copy_image_to_label": False,
    # default augmentation probabilities
    "additive_noise_p_per_sample": 0.2,
    "biasfield_p_per_sample": 0.0,
    "blurring_p_per_sample": 0.2,
    "blurring_p_per_channel": 0.5,
    "elastic_deform_p_per_sample": 0.33,
    "gamma_p_per_sample": 0.2,
    "gamma_p_invert_image": 0.05,
    "gibbs_ringing_p_per_sample": 0.0,
    "mirror_p_per_sample": 0.0,
    "mirror_p_per_axis": 0.33,
    "motion_ghosting_p_per_sample": 0.0,
    "multiplicative_noise_p_per_sample": 0.2,
    "rotation_p_per_sample": 0.2,
    "rotation_p_per_axis": 0.66,
    "scale_p_per_sample": 0.2,
    "simulate_lowres_p_per_sample": 0.2,
    "simulate_lowres_p_per_channel": 0.5,
    "simulate_lowres_p_per_axis": 0.33,
    # default augmentation values
}



from yucca.data.augmentation.transforms.YuccaTransform import YuccaTransform
import numpy as np
from typing import Tuple
from yucca.functional.transforms import blur


class Blur(YuccaTransform):
    def __init__(
        self,
        data_key="image",
        p_per_sample: float = 1.0,
        p_per_channel: float = 0.5,
        sigma=(0.5, 1.0),
        clip_to_input_range=False,
    ):
        self.data_key = data_key
        self.p_per_sample = p_per_sample
        self.p_per_channel = p_per_channel
        self.sigma = sigma
        self.clip_to_input_range = clip_to_input_range

    @staticmethod
    def get_params(sigma: Tuple[float]):
        sigma = np.random.uniform(*sigma)
        return sigma

    def __blur__(self, image, sigma):
        for c in range(image.shape[0]):
            if np.random.uniform() < self.p_per_channel:
                image[c] = blur(image[c], sigma, clip_to_input_range=self.clip_to_input_range)
        return image

    def __call__(self, packed_data_dict=None, **unpacked_data_dict):
        data_dict = packed_data_dict if packed_data_dict else unpacked_data_dict
        assert (
            len(data_dict[self.data_key].shape) == 5 or len(data_dict[self.data_key].shape) == 4
        ), f"Incorrect data size or shape.\
            \nShould be (b, c, x, y, z) or (b, c, x, y) and is: {data_dict[self.data_key].shape}"

        for b in range(data_dict[self.data_key].shape[0]):
            if np.random.uniform() < self.p_per_sample:
                sigma = self.get_params(self.sigma)
                data_dict[self.data_key][b] = self.__blur__(data_dict[self.data_key][b], sigma)
        return data_dict



import torch
import torch.nn as nn
from yucca.optimization.loss_functions.CE import BCEWithLogitsLoss
from yucca.optimization.loss_functions.Dice import SoftSigmoidDiceLoss


class SigmoidDiceBCE(nn.Module):

    def __init__(
        self,
        sigmoid_dice_kwargs={},
        ce_kwargs={},
        weight_ce=1,
        weight_dice=1,
        log_dice=False,
        ignore_label=None,
    ):
        """
        CAREFUL. Weights for CE and Dice do not need to sum to one. You can set whatever you want.
        :param soft_dice_kwargs:
        :param ce_kwargs:
        :param aggregate:
        :param square_dice:
        :param weight_ce:
        :param weight_dice:
        """
        super(SigmoidDiceBCE, self).__init__()
        if ignore_label is not None:
            ce_kwargs["reduction"] = "none"
        self.log_dice = log_dice
        self.weight_dice = weight_dice
        self.weight_ce = weight_ce
        self.ce = BCEWithLogitsLoss()

        self.ignore_label = ignore_label

        self.dc = SoftSigmoidDiceLoss(**sigmoid_dice_kwargs)

    def forward(self, net_output, target):
        """
        target must be b, c, x, y(, z) with c=1
        :param net_output:
        :param target:
        :return:
        """
        if self.ignore_label is not None:
            assert target.shape[1] == 1, "not implemented for one hot encoding"
            mask = target != self.ignore_label
            target[~mask] = 0
            mask = mask.float()
        else:
            mask = None

        dc_loss = self.dc(net_output, target, loss_mask=mask) if self.weight_dice != 0 else 0
        if self.log_dice:
            dc_loss = -torch.log(-dc_loss)

        ce_loss = self.ce(net_output, target) if self.weight_ce != 0 else 0
        if self.ignore_label is not None:
            ce_loss *= mask[:, 0]
            ce_loss = ce_loss.sum() / mask.sum()

        result = self.weight_ce * ce_loss + self.weight_dice * dc_loss

        return result



from yucca.functional.transforms.blur import blur
from yucca.functional.transforms.bias_field import bias_field
from yucca.functional.transforms.gamma import augment_gamma
from yucca.functional.transforms.motion_ghosting import motion_ghosting
from yucca.functional.transforms.noise import additive_noise, multiplicative_noise
from yucca.functional.transforms.ringing import gibbs_ringing
from yucca.functional.transforms.sampling import downsample_label, simulate_lowres
from yucca.functional.transforms.masking import mask_batch
from yucca.functional.transforms.spatial import spatial



from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p, subfiles
from yucca.pipeline.task_conversion.utils import generate_dataset_json
import shutil
import gzip
from yucca.paths import yucca_raw_data


def convert(path: str, subdir: str = "HarP"):
    # INPUT DATA
    path = f"{path}/{subdir}"
    file_suffix = ".nii"

    # Train/Test Splits
    labels_dir = join(path, "Labels")
    images_dir = join(path, "Images")

    # 2 to 99 are used as training samples
    # 100 to 137 are used as test samples
    all_samples = subfiles(labels_dir, join=False, suffix=file_suffix)
    tr_ids = ["ADNI_" + f"{id:03}" for id in range(2, 99)]
    training_samples = [sample for sample in all_samples if sample[:8] in tr_ids]
    ts_ids = ["ADNI_" + f"{id:03}" for id in range(100, 137)]
    test_samples = [sample for sample in all_samples if sample[:8] in ts_ids]

    ###OUTPUT DATA
    # Target names
    task_name = "Task004_HarP"
    prefix = "HarP"

    # Target paths
    target_base = join(yucca_raw_data, task_name)

    target_imagesTr = join(target_base, "imagesTr")
    target_labelsTr = join(target_base, "labelsTr")

    target_imagesTs = join(target_base, "imagesTs")
    target_labelsTs = join(target_base, "labelsTs")

    maybe_mkdir_p(target_imagesTr)
    maybe_mkdir_p(target_labelsTs)
    maybe_mkdir_p(target_imagesTs)
    maybe_mkdir_p(target_labelsTr)

    ###Populate Target Directory###
    # This is likely also the place to apply any re-orientation, resampling and/or label correction.
    for sTr in training_samples:
        serial_number = sTr[:-4]
        image_file = open(join(images_dir, sTr), "rb")
        label = open(join(labels_dir, sTr), "rb")
        shutil.copyfileobj(image_file, gzip.open(f"{target_imagesTr}/{prefix}_{serial_number}_000.nii.gz", "wb"))
        shutil.copyfileobj(label, gzip.open(f"{target_labelsTr}/{prefix}_{serial_number}.nii.gz", "wb"))

    for sTs in test_samples:
        serial_number = sTs[:-4]
        image_file = open(join(images_dir, sTs), "rb")
        label = open(join(labels_dir, sTs), "rb")
        shutil.copyfileobj(image_file, gzip.open(f"{target_imagesTs}/{prefix}_{serial_number}_000.nii.gz", "wb"))
        shutil.copyfileobj(label, gzip.open(f"{target_labelsTs}/{prefix}_{serial_number}.nii.gz", "wb"))

    generate_dataset_json(
        join(target_base, "dataset.json"),
        target_imagesTr,
        target_imagesTs,
        ("T1",),
        labels={0: "background", 1: "Left Hippocampus", 2: "Right Hippocampus"},
        dataset_name=task_name,
        license="hands off!",
        dataset_description="HarP Dataset",
        dataset_reference="",
    )


if __name__ == "__main__":
    convert()



from .dict import without_keys
from .files_and_folders import (
    recursive_find_python_class,
    recursive_find_realpath,
    recursive_rename,
    rename_file_or_dir,
    replace_in_file,
)
from .kwargs import filter_kwargs, getattr_for_kwargs
from .loading import load_yaml, read_file_to_nifti_or_np
from .nib_utils import get_nib_orientation, get_nib_spacing, reorient_nib_image
from .saving import (
    merge_softmax_from_folders,
    save_nifti_from_numpy,
    save_png_from_numpy,
    save_prediction_from_logits,
    save_txt_from_numpy,
)
from .softmax import softmax
from .torch_utils import maybe_to_gpu, get_available_device, flush_and_get_torch_memory_allocated, measure_FLOPs
from .type_conversions import np_to_nifti_with_empty_header, nifti_or_np_to_np, png_to_nifti



from .confusion_matrix import torch_confusion_matrix_from_logits, torch_get_tp_fp_tn_fn
from .metrics import (
    dice,
    dice_per_label,
    jaccard,
    jaccard_per_label,
    sensitivity,
    specificity,
    precision,
    volume_similarity,
    f1,
    accuracy,
    auroc,
    TP,
    FP,
    FN,
    total_pos_gt,
    total_pos_pred,
)
from .obj_metrics import get_obj_stats_for_label, obj_get_tp_fp_fn_gtvols_predvols
from .surface_metrics import get_surface_metrics_for_label



import torch
from yucca.pipeline.preprocessing.YuccaPreprocessor import YuccaPreprocessor


class ClassificationPreprocessor(YuccaPreprocessor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # set up for classification
        self.classification = True
        self.label_exists = True
        self.preprocess_label = False

    def reverse_preprocessing(self, images: torch.Tensor, image_properties: dict):
        """
        Expected shape of images are:
        (b, c, x)
        """
        image_properties["save_format"] = "txt"
        return images.cpu().numpy(), image_properties



import nibabel as nib
from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p, subfiles
from yucca.pipeline.task_conversion.utils import generate_dataset_json
from yucca.paths import yucca_raw_data, yucca_source


def convert(path: str = yucca_source, subdir: str = "decathlon", subsubdir: str = "Task01_BrainTumour"):
    # INPUT DATA
    path = f"{path}/{subdir}/{subsubdir}"

    file_suffix = ".nii.gz"

    # Train/Test Splits
    images_dir_tr = join(path, "imagesTr")
    labels_dir_tr = join(path, "labelsTr")
    images_dir_ts = join(path, "imagesTs")

    ###OUTPUT DATA
    # Target names
    task_name = "Task031_MSD_BrainTumour"

    # Target paths
    target_base = join(yucca_raw_data, task_name)

    target_imagesTr = join(target_base, "imagesTr")
    target_labelsTr = join(target_base, "labelsTr")

    target_imagesTs = join(target_base, "imagesTs")
    target_labelsTs = join(target_base, "labelsTs")

    maybe_mkdir_p(target_imagesTr)
    maybe_mkdir_p(target_labelsTs)
    maybe_mkdir_p(target_imagesTs)
    maybe_mkdir_p(target_labelsTr)

    for sTr in subfiles(labels_dir_tr, join=False):
        image = nib.load(join(images_dir_tr, sTr))
        label = nib.load(join(labels_dir_tr, sTr))
        sTr = sTr[: -len(file_suffix)]

        # I believe this is the order they're saved in but as the model wont care I also wont care too much
        # If this is important I suggest you double check
        flair = image.slicer[:, :, :, 0]
        t1w = image.slicer[:, :, :, 1]
        t1gd = image.slicer[:, :, :, 2]
        t2w = image.slicer[:, :, :, 3]

        nib.save(flair, f"{target_imagesTr}/{sTr}_000.nii.gz")
        nib.save(t1w, f"{target_imagesTr}/{sTr}_001.nii.gz")
        nib.save(t1gd, f"{target_imagesTr}/{sTr}_002.nii.gz")
        nib.save(t2w, f"{target_imagesTr}/{sTr}_003.nii.gz")

        nib.save(label, f"{target_labelsTr}/{sTr}.nii.gz")

    for sTs in subfiles(images_dir_ts, join=False):
        image = nib.load(join(images_dir_ts, sTs))
        sTs = sTs[: -len(file_suffix)]

        flair = image.slicer[:, :, :, 0]
        t1w = image.slicer[:, :, :, 1]
        t1gd = image.slicer[:, :, :, 2]
        t2w = image.slicer[:, :, :, 3]

        nib.save(flair, f"{target_imagesTs}/{sTs}_000.nii.gz")
        nib.save(t1w, f"{target_imagesTs}/{sTs}_001.nii.gz")
        nib.save(t1gd, f"{target_imagesTs}/{sTs}_002.nii.gz")
        nib.save(t2w, f"{target_imagesTs}/{sTs}_003.nii.gz")

    generate_dataset_json(
        join(target_base, "dataset.json"),
        target_imagesTr,
        target_imagesTs,
        modalities=("FLAIR", "T1w", "t1gd", "T2W"),
        labels={0: "background", 1: "edema", 2: "non-enhancing tumor", 3: "enhancing tumour"},
        dataset_name=task_name,
        license="CC-BY-SA 4.0",
        dataset_description="Decathlon: Brain Tumour",
        dataset_reference="King's College London",
    )


if __name__ == "__main__":
    convert()



import numpy as np
import cc3d
from yucca.functional.utils.nib_utils import get_nib_spacing
from yucca.functional.evaluation.metrics import sensitivity, precision, f1


def get_obj_stats_for_label(gt, pred, label, as_binary=False):
    spacing = get_nib_spacing(gt)
    gt = np.around(gt.get_fdata())
    pred = np.around(pred.get_fdata())
    if as_binary:
        gt = gt.astype(bool).astype(int)
        pred = pred.astype(bool).astype(int)
    labeldict = {}
    if label == 0:
        labeldict["_OBJ Total Objects Prediction"] = 0
        labeldict["_OBJ Total Objects Ground Truth"] = 0
        labeldict["_OBJ True Positives"] = 0
        labeldict["_OBJ False Positives"] = 0
        labeldict["_OBJ False Negatives"] = 0
        labeldict["_OBJ Mean Volume Prediction"] = 0
        labeldict["_OBJ Mean Volume Ground Truth"] = 0
        labeldict["_OBJ sensitivity"] = 0
        labeldict["_OBJ precision"] = 0
        labeldict["_OBJ F1"] = 0
    else:
        pred_for_label = np.where(pred == label, 1, 0).astype(bool)
        gt_for_label = np.where(gt == label, 1, 0).astype(bool)
        cc_gt, cc_gt_n = cc3d.connected_components(gt_for_label, connectivity=26, return_N=True)
        cc_pred, cc_pred_n = cc3d.connected_components(pred_for_label, connectivity=26, return_N=True)
        tp, fp, fn, gtvols, predvols = obj_get_tp_fp_fn_gtvols_predvols(cc_gt, cc_pred, cc_gt_n, cc_pred_n)

        labeldict["_OBJ Total Objects Prediction"] = cc_pred_n
        labeldict["_OBJ Total Objects Ground Truth"] = cc_gt_n
        labeldict["_OBJ True Positives"] = tp
        labeldict["_OBJ False Positives"] = fp
        labeldict["_OBJ False Negatives"] = fn
        labeldict["_OBJ Mean Volume Prediction"] = float(np.prod(spacing) * np.mean(predvols))
        labeldict["_OBJ Mean Volume Ground Truth"] = float(np.prod(spacing) * np.mean(gtvols))
        labeldict["_OBJ sensitivity"] = sensitivity(tp, fp, 0, fn)
        labeldict["_OBJ precision"] = precision(tp, fp, 0, fn)
        labeldict["_OBJ F1"] = f1(tp, fp, 0, fn)
    return labeldict


def obj_get_tp_fp_fn_gtvols_predvols(cc_gt, cc_pred, n_cc_gt, n_cc_pred):
    tp = 0
    fp = 0
    fn = 0
    predvols = []
    gtvols = []
    if n_cc_gt > 0:
        for _, binary_cluster_image in cc3d.each(cc_gt, binary=True, in_place=True):
            gtvols.append(binary_cluster_image.sum())
            if np.logical_and(binary_cluster_image, cc_pred).any():
                tp += 1
            else:
                fn += 1
    if n_cc_pred > 0:
        for _, binary_cluster_image in cc3d.each(cc_pred, binary=True, in_place=True):
            predvols.append(binary_cluster_image.sum())
            if not np.logical_and(binary_cluster_image, cc_gt).any():
                fp += 1
    if not predvols:
        predvols = 0
    if not gtvols:
        gtvols = 0
    return tp, fp, fn, gtvols, predvols


def n_cc_gt(cc_gt, cc_pred, n_cc_gt, n_cc_pred):  # noqa: U100
    return n_cc_gt


def n_cc_pred(cc_gt, cc_pred, n_cc_gt, n_cc_pred):  # noqa: U100
    return n_cc_pred



"""
Mean Squared Error loss wrapper
Makes loss function robust to target tensors with a channel dimension
"""

from torch import nn, Tensor


class MSE(nn.MSELoss):
    def __init__(self):
        super().__init__()

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        return super().forward(input, target)



import torch
import lightning as L
import datetime
from dataclasses import dataclass
from typing import Union


@dataclass
class SeedConfig:
    seed: int

    def lm_hparams(self):
        return {"seed": self.seed}


def seed_everything_and_get_seed_config(manual_seed: Union[int, None] = None, ckpt_seed: Union[int, None] = None):
    # Priority of seeds.
    # 1. If manual seed is specified we use that
    # 2. If seed is found in checkpoint continue with that (e.g. continued training or finetuning)
    # 3. If no seed is found we generate one from the datetime object
    if manual_seed is not None:
        seed = manual_seed
    elif ckpt_seed is not None:
        seed = ckpt_seed
    else:
        dt = datetime.datetime.now()
        seq = int(dt.strftime("%m%d%H%M%S"))
        seed = seq

    L.seed_everything(seed=seed, workers=True)
    seed = torch.initial_seed()

    return SeedConfig(
        seed=seed,
    )



import shutil
from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p, subfiles
from yucca.task_conversion.utils import generate_dataset_json, remove_punctuation_and_spaces
from yucca.paths import yucca_raw_data
from sklearn.model_selection import train_test_split
from tqdm import tqdm


def convert(path: str, subdir: str = "Autopet"):
    """INPUT DATA - Define input path and suffixes"""
    path = join(path, subdir)

    """ OUTPUT DATA - Define the task name and prefix """
    task_name = "Task013_AutoPET3"
    task_prefix = "AutoPET3"

    # Target paths
    target_base = join(yucca_raw_data, task_name)

    target_imagesTr = join(target_base, "imagesTr")
    target_labelsTr = join(target_base, "labelsTr")

    target_imagesTs = join(target_base, "imagesTs")
    target_labelsTs = join(target_base, "labelsTs")

    maybe_mkdir_p(target_imagesTr)
    maybe_mkdir_p(target_labelsTs)
    maybe_mkdir_p(target_imagesTs)
    maybe_mkdir_p(target_labelsTr)

    # Input paths
    images_dir = join(path, "imagesTr")
    labels_dir = join(path, "labelsTr")

    training_samples, test_samples = train_test_split(subfiles(labels_dir, join=False), random_state=859032)

    ###Populate Target Directory###
    for sTr in tqdm(training_samples, desc="Training cases"):
        sTr = sTr[: -len(".nii.gz")]
        sTr_valid = remove_punctuation_and_spaces(sTr)

        # Loading relevant modalities and the ground truth
        src_image_file_path1 = join(images_dir, sTr + "_0000.nii.gz")
        src_image_file_path2 = join(images_dir, sTr + "_0001.nii.gz")
        src_label_path = join(labels_dir, sTr + ".nii.gz")

        dst_image_file_path1 = f"{target_imagesTr}/{task_prefix}_{sTr_valid}_000.nii.gz"
        dst_image_file_path2 = f"{target_imagesTr}/{task_prefix}_{sTr_valid}_001.nii.gz"
        dst_label_path = f"{target_labelsTr}/{task_prefix}_{sTr_valid}.nii.gz"

        shutil.copy2(src_image_file_path1, dst_image_file_path1)
        shutil.copy2(src_image_file_path2, dst_image_file_path2)
        shutil.copy2(src_label_path, dst_label_path)

    for sTs in tqdm(test_samples, desc="Testing cases"):
        sTs = sTs[: -len(".nii.gz")]
        sTs_valid = remove_punctuation_and_spaces(sTs)

        # Loading relevant modalities and the ground truth
        src_image_file_path1 = join(images_dir, sTs + "_0000.nii.gz")
        src_image_file_path2 = join(images_dir, sTs + "_0001.nii.gz")
        src_label_path = join(labels_dir, sTs + ".nii.gz")

        dst_image_file_path1 = f"{target_imagesTs}/{task_prefix}_{sTs_valid}_000.nii.gz"
        dst_image_file_path2 = f"{target_imagesTs}/{task_prefix}_{sTs_valid}_001.nii.gz"
        dst_label_path = f"{target_labelsTs}/{task_prefix}_{sTs_valid}.nii.gz"

        shutil.copy2(src_image_file_path1, dst_image_file_path1)
        shutil.copy2(src_image_file_path2, dst_image_file_path2)
        shutil.copy2(src_label_path, dst_label_path)

    generate_dataset_json(
        join(target_base, "dataset.json"),
        target_imagesTr,
        target_imagesTs,
        ("PET", "CT"),
        labels={0: "background", 1: "tumour"},
        dataset_name=task_name,
        license="CC BY-NC 4.0 DEED",
        dataset_description="AutoPET3 Multicenter Multitracer Generalization",
        dataset_reference="https://autopet-iii.grand-challenge.org/",
    )



"""
Unified Focal losses from the paper:

Yeung M, Sala E, Schönlieb CB, Rundo L. 
    Unified Focal loss: Generalising Dice and cross entropy-based losses to handle class imbalanced 
    medical image segmentation. 
    Comput Med Imaging Graph. 2022 Jan;95:102026. 
    doi: 10.1016/j.compmedimag.2021.102026. 
    Epub 2021 Dec 13. PMID: 34953431; PMCID: PMC8785124.

Code implementation from:
https://github.com/oikosohn/compound-loss-pytorch

"""

import torch
import torch.nn as nn


# Helper function to enable loss function to be flexibly used for
# both 2D or 3D image segmentation - source: https://github.com/frankkramer-lab/MIScnn


def identify_axis(shape):
    # Three dimensional
    if len(shape) == 5:
        return [2, 3, 4]

    # Two dimensional
    elif len(shape) == 4:
        return [2, 3]

    # Exception - Unknown
    else:
        raise ValueError("Metric: Shape of tensor is neither 2D or 3D.")


class SymmetricFocalLoss(nn.Module):
    """
    Parameters
    ----------
    delta : float, optional
        controls weight given to false positive and false negatives, by default 0.7
    gamma : float, optional
        Focal Tversky loss' focal parameter controls degree of down-weighting of easy examples, by default 2.0
    epsilon : float, optional
        clip values to prevent division by zero error
    """

    def __init__(self, delta=0.7, gamma=2.0, epsilon=1e-07):
        super(SymmetricFocalLoss, self).__init__()
        self.delta = delta
        self.gamma = gamma
        self.epsilon = epsilon

    def forward(self, y_pred, y_true):
        y_pred = torch.clamp(y_pred, self.epsilon, 1.0 - self.epsilon)
        cross_entropy = -y_true * torch.log(y_pred)

        # Calculate losses separately for each class
        back_ce = torch.pow(1 - y_pred[:, 0, :, :], self.gamma) * cross_entropy[:, 0, :, :]
        back_ce = (1 - self.delta) * back_ce

        fore_ce = torch.pow(1 - y_pred[:, 1, :, :], self.gamma) * cross_entropy[:, 1, :, :]
        fore_ce = self.delta * fore_ce

        loss = torch.mean(torch.sum(torch.stack([back_ce, fore_ce], axis=-1), axis=-1))

        return loss


class AsymmetricFocalLoss(nn.Module):
    """For Imbalanced datasets
    Parameters
    ----------
    delta : float, optional
        controls weight given to false positive and false negatives, by default 0.25
    gamma : float, optional
        Focal Tversky loss' focal parameter controls degree of down-weighting of easy examples, by default 2.0
    epsilon : float, optional
        clip values to prevent division by zero error
    """

    def __init__(self, delta=0.7, gamma=2.0, epsilon=1e-07):
        super(AsymmetricFocalLoss, self).__init__()
        self.delta = delta
        self.gamma = gamma
        self.epsilon = epsilon

    def forward(self, y_pred, y_true):
        y_pred = torch.clamp(y_pred, self.epsilon, 1.0 - self.epsilon)
        cross_entropy = -y_true * torch.log(y_pred)

        # Calculate losses separately for each class, only suppressing background class
        back_ce = torch.pow(1 - y_pred[:, 0, :, :], self.gamma) * cross_entropy[:, 0, :, :]
        back_ce = (1 - self.delta) * back_ce

        fore_ce = cross_entropy[:, 1, :, :]
        fore_ce = self.delta * fore_ce

        loss = torch.mean(torch.sum(torch.stack([back_ce, fore_ce], axis=-1), axis=-1))

        return loss


class SymmetricFocalTverskyLoss(nn.Module):
    """This is the implementation for binary segmentation.
    Parameters
    ----------
    delta : float, optional
        controls weight given to false positive and false negatives, by default 0.7
    gamma : float, optional
        focal parameter controls degree of down-weighting of easy examples, by default 0.75
    smooth : float, optional
        smooithing constant to prevent division by 0 errors, by default 0.000001
    epsilon : float, optional
        clip values to prevent division by zero error
    """

    def __init__(self, delta=0.7, gamma=0.75, epsilon=1e-07):
        super(SymmetricFocalTverskyLoss, self).__init__()
        self.delta = delta
        self.gamma = gamma
        self.epsilon = epsilon

    def forward(self, y_pred, y_true):
        y_pred = torch.clamp(y_pred, self.epsilon, 1.0 - self.epsilon)
        axis = identify_axis(y_true.size())

        # Calculate true positives (tp), false negatives (fn) and false positives (fp)
        tp = torch.sum(y_true * y_pred, axis=axis)
        fn = torch.sum(y_true * (1 - y_pred), axis=axis)
        fp = torch.sum((1 - y_true) * y_pred, axis=axis)
        dice_class = (tp + self.epsilon) / (tp + self.delta * fn + (1 - self.delta) * fp + self.epsilon)

        # Calculate losses separately for each class, enhancing both classes
        back_dice = (1 - dice_class[:, 0]) * torch.pow(1 - dice_class[:, 0], -self.gamma)
        fore_dice = (1 - dice_class[:, 1]) * torch.pow(1 - dice_class[:, 1], -self.gamma)

        # Average class scores
        loss = torch.mean(torch.stack([back_dice, fore_dice], axis=-1))
        return loss


class AsymmetricFocalTverskyLoss(nn.Module):
    """This is the implementation for binary segmentation.
    Parameters
    ----------
    delta : float, optional
        controls weight given to false positive and false negatives, by default 0.7
    gamma : float, optional
        focal parameter controls degree of down-weighting of easy examples, by default 0.75
    smooth : float, optional
        smooithing constant to prevent division by 0 errors, by default 0.000001
    epsilon : float, optional
        clip values to prevent division by zero error
    """

    def __init__(self, delta=0.7, gamma=0.75, epsilon=1e-07):
        super(AsymmetricFocalTverskyLoss, self).__init__()
        self.delta = delta
        self.gamma = gamma
        self.epsilon = epsilon

    def forward(self, y_pred, y_true):
        # Clip values to prevent division by zero error
        y_pred = torch.clamp(y_pred, self.epsilon, 1.0 - self.epsilon)
        axis = identify_axis(y_true.size())

        # Calculate true positives (tp), false negatives (fn) and false positives (fp)
        print(y_true.shape, y_pred.shape)
        tp = torch.sum(y_true * y_pred, axis=axis)
        fn = torch.sum(y_true * (1 - y_pred), axis=axis)
        fp = torch.sum((1 - y_true) * y_pred, axis=axis)
        dice_class = (tp + self.epsilon) / (tp + self.delta * fn + (1 - self.delta) * fp + self.epsilon)

        # Calculate losses separately for each class, only enhancing foreground class
        back_dice = 1 - dice_class[:, 0]
        fore_dice = (1 - dice_class[:, 1]) * torch.pow(1 - dice_class[:, 1], -self.gamma)

        # Average class scores
        loss = torch.mean(torch.stack([back_dice, fore_dice], axis=-1))
        return loss


class SymmetricUnifiedFocalLoss(nn.Module):
    """The Unified Focal loss is a new compound loss function that unifies Dice-based and cross entropy-based loss functions into a single framework.
    Parameters
    ----------
    weight : float, optional
        represents lambda parameter and controls weight given to symmetric Focal Tversky loss and symmetric Focal loss, by default 0.5
    delta : float, optional
        controls weight given to each class, by default 0.6
    gamma : float, optional
        focal parameter controls the degree of background suppression and foreground enhancement, by default 0.5
    epsilon : float, optional
        clip values to prevent division by zero error
    """

    def __init__(self, weight=0.5, delta=0.6, gamma=0.5):
        super(SymmetricUnifiedFocalLoss, self).__init__()
        self.weight = weight
        self.delta = delta
        self.gamma = gamma

    def forward(self, y_pred, y_true):
        symmetric_ftl = SymmetricFocalTverskyLoss(delta=self.delta, gamma=self.gamma)(y_pred, y_true)
        symmetric_fl = SymmetricFocalLoss(delta=self.delta, gamma=self.gamma)(y_pred, y_true)
        if self.weight is not None:
            return (self.weight * symmetric_ftl) + ((1 - self.weight) * symmetric_fl)
        else:
            return symmetric_ftl + symmetric_fl


class AsymmetricUnifiedFocalLoss(nn.Module):
    """The Unified Focal loss is a new compound loss function that unifies Dice-based and cross entropy-based loss functions into a single framework.
    Parameters
    ----------
    weight : float, optional
        represents lambda parameter and controls weight given to asymmetric Focal Tversky loss and asymmetric Focal loss, by default 0.5
    delta : float, optional
        controls weight given to each class, by default 0.6
    gamma : float, optional
        focal parameter controls the degree of background suppression and foreground enhancement, by default 0.5
    epsilon : float, optional
        clip values to prevent division by zero error
    """

    def __init__(self, weight=0.5, delta=0.6, gamma=0.2):
        super(AsymmetricUnifiedFocalLoss, self).__init__()
        self.weight = weight
        self.delta = delta
        self.gamma = gamma

    def forward(self, y_pred, y_true):
        # Obtain Asymmetric Focal Tversky loss
        asymmetric_ftl = AsymmetricFocalTverskyLoss(delta=self.delta, gamma=self.gamma)(y_pred, y_true)

        # Obtain Asymmetric Focal loss
        asymmetric_fl = AsymmetricFocalLoss(delta=self.delta, gamma=self.gamma)(y_pred, y_true)

        # Return weighted sum of Asymmetrical Focal loss and Asymmetric Focal Tversky loss
        if self.weight is not None:
            return (self.weight * asymmetric_ftl) + ((1 - self.weight) * asymmetric_fl)
        else:
            return asymmetric_ftl + asymmetric_fl



import numpy as np
import torch
import os
import logging
from typing import Union, Literal, Optional
from batchgenerators.utilities.file_and_folder_operations import subfiles, load_pickle, isfile
from yucca.data.augmentation.transforms.cropping_and_padding import CropPad
from yucca.data.augmentation.transforms.formatting import NumpyToTorch


class YuccaTrainDataset(torch.utils.data.Dataset):
    def __init__(
        self,
        samples: list,
        patch_size: list | tuple,
        keep_in_ram: Union[bool, None] = None,
        label_dtype: Optional[Union[int, float]] = None,
        composed_transforms=None,
        task_type: Literal["classification", "segmentation", "self-supervised", "contrastive"] = "segmentation",
        allow_missing_modalities=False,
    ):
        self.all_cases = samples
        self.composed_transforms = composed_transforms
        self.patch_size = patch_size
        self.task_type = task_type
        assert task_type in ["classification", "segmentation", "self-supervised", "contrastive"]
        self.supervised = self.task_type in ["classification", "segmentation"]
        self.label_dtype = label_dtype
        self.allow_missing_modalities = allow_missing_modalities

        self.already_loaded_cases = {}

        # for segmentation and classification we override the default None
        # because arrays are saved as floats and we want them to be ints.
        if self.label_dtype is None:
            if self.supervised:
                self.label_dtype = torch.int32

        self.croppad = CropPad(patch_size=self.patch_size, p_oversample_foreground=0.33)
        self.to_torch = NumpyToTorch(label_dtype=self.label_dtype)

        self._keep_in_ram = keep_in_ram

    @property
    def keep_in_ram(self):
        if self._keep_in_ram is not None:
            return self._keep_in_ram
        if len(self.all_cases) < 50:
            self._keep_in_ram = True
        else:
            logging.debug("Large dataset detected. Will not keep cases in RAM during training.")
            self._keep_in_ram = False
        return self._keep_in_ram

    def load_and_maybe_keep_pickle(self, path):
        path = path + ".pkl"
        if not self.keep_in_ram:
            return load_pickle(path)
        if path in self.already_loaded_cases:
            return self.already_loaded_cases[path]
        self.already_loaded_cases[path] = load_pickle(path)
        return self.already_loaded_cases[path]

    def load_and_maybe_keep_volume(self, path):
        path = path + ".npy"
        if not self.keep_in_ram:
            if isfile(path):
                try:
                    return np.load(path, "r")
                except ValueError:
                    return np.load(path, allow_pickle=True)
            else:
                print("uncompressed data was not found.")

        if isfile(path):
            if path in self.already_loaded_cases:
                return self.already_loaded_cases[path]
            try:
                self.already_loaded_cases[path] = np.load(path, "r")
            except ValueError:
                self.already_loaded_cases[path] = np.load(path, allow_pickle=True)
        else:
            print("uncompressed data was not found.")

        return self.already_loaded_cases[path]

    def __len__(self):
        return len(self.all_cases)

    def __getitem__(self, idx):
        # remove extension if file splits include extensions
        case, _ = os.path.splitext(self.all_cases[idx])
        data = self.load_and_maybe_keep_volume(case)
        metadata = self.load_and_maybe_keep_pickle(case)

        if self.allow_missing_modalities:
            image, label = self.unpack_with_zeros(data, supervised=self.supervised)
        else:
            image, label = self.unpack(data, supervised=self.supervised)

        data_dict = {"file_path": case}  # metadata that can be very useful for debugging.
        if self.task_type in ["classification", "segmentation"]:
            data_dict.update({"image": image, "label": label})
        elif self.task_type == "self-supervised":
            data_dict.update({"image": image})
        elif self.task_type == "contrastive":
            view1 = self._transform({"image": image}, case)["image"]
            view2 = self._transform({"image": image}, case)["image"]
            data_dict.update({"image": (view1, view2)})
            return data_dict
        else:
            logging.error(f"Task Type not recognized. Found {self.task_type}")
        return self._transform(data_dict, metadata)

    def _transform(self, data_dict, metadata):
        data_dict = self.croppad(data_dict, metadata)
        if self.composed_transforms is not None:
            data_dict = self.composed_transforms(data_dict)
        return self.to_torch(data_dict)

    def unpack(self, data, supervised: bool):
        if supervised:
            return data[:-1], data[-1:]
        return data, None

    def unpack_with_zeros(self, data, supervised: bool):
        assert data.dtype == "object", "allow missing modalities is true but dtype is not object"

        # First find the array with the largest array.
        # in classification this avoids setting the zero array to the 1d array with classes
        sizes = [i.size for i in data]
        idx_largest_array = np.where(sizes == np.max(sizes))[0][0]

        # replace missing modalities with zero-filed arrays
        for idx, i in enumerate(data):
            if i.size == 0:
                data[idx] = np.zeros(data[idx_largest_array].squeeze().shape)

        # unpack array into images and (maybe) labels
        if supervised:
            images = np.array([mod for mod in data[:-1]])
            label = data[-1:][0]
        else:
            images = np.array([mod for mod in data])
            label = None
        return images, label


class YuccaTestDataset(torch.utils.data.Dataset):
    def __init__(
        self,
        raw_data_dir: str,
        pred_save_dir: str,
        overwrite_predictions: bool = False,
        suffix="nii.gz",
        pred_include_cases: list = None,
    ):
        self.data_path = raw_data_dir
        self.pred_save_dir = pred_save_dir
        self.overwrite = overwrite_predictions
        self.suffix = suffix
        self.pred_include_cases = pred_include_cases
        self.unique_cases = np.unique(
            [i[: -len("_000." + suffix)] for i in subfiles(self.data_path, suffix=self.suffix, join=False)]
        )
        assert len(self.unique_cases) > 0, f"No cases found in {self.data_path}. Looking for files with suffix: {self.suffix}"

        self.cases_already_predicted = np.unique(
            [i[: -len("." + suffix)] for i in subfiles(self.pred_save_dir, suffix=self.suffix, join=False)]
        )
        logging.info(f"Found {len(self.cases_already_predicted)} already predicted cases. Overwrite: {self.overwrite}")
        if not self.overwrite:
            self.unique_cases = [i for i in self.unique_cases if i not in self.cases_already_predicted]
        if self.pred_include_cases is not None:
            self.unique_cases = [i for i in self.unique_cases if i in self.pred_include_cases]

    def __len__(self):
        return len(self.unique_cases)

    def __getitem__(self, idx):
        # Here we generate the paths to the cases along with their ID which they will be saved as.
        # we pass "case" as a list of strings and case_id as a string to the dataloader which
        # will convert them to a list of tuples of strings and a tuple of a string.
        # i.e. ['path1', 'path2'] -> [('path1',), ('path2',)]
        case_id = self.unique_cases[idx]
        case = [
            impath
            for impath in subfiles(self.data_path, suffix=self.suffix)
            if os.path.split(impath)[-1][: -len("_000." + self.suffix)] == case_id
        ]
        return case, case_id


if __name__ == "__main__":
    import torch
    from yucca.paths import yucca_preprocessed_data
    from batchgenerators.utilities.file_and_folder_operations import join
    from yucca.data.samplers import InfiniteRandomSampler

    files = subfiles(join(yucca_preprocessed_data, "Task001_OASIS/YuccaPlanner"), suffix="npy")
    ds = YuccaTrainDataset(files, patch_size=(12, 12, 12))
    sampler = InfiniteRandomSampler(ds)
    dl = torch.utils.data.DataLoader(ds, num_workers=2, batch_size=2, sampler=sampler)



import torch
import numpy as np


class RandomBatchSampler(torch.utils.data.Sampler):
    def __init__(self, dataset: torch.utils.data.Dataset, batch_size: int = None):
        assert len(dataset) > 0
        self.dataset = dataset
        self.batch_size = batch_size

    def __iter__(self):
        yield np.random.choice(len(self.dataset), self.batch_size)


class InfiniteRandomSampler(torch.utils.data.Sampler):
    """Return random indices from [0-n) infinitely.

    Arguments:
        dset_size (int): Size of the dataset to sample.
    """

    def __init__(self, dataset: torch.utils.data.Dataset):
        self.dset_size = len(dataset)

    def __iter__(self):
        # Create a random number generator (optional, makes the sampling independent of the base RNG)
        rng = torch.Generator()
        seed = torch.empty((), dtype=torch.int64).random_().item()
        rng.manual_seed(seed)
        return _infinite_generator(self.dset_size, rng)

    def __len__(self):
        return float("inf")


def _infinite_generator(n, rng):
    """Inifinitely returns a number in [0, n)."""
    while True:
        yield from torch.randperm(n, generator=rng).tolist()



import numpy as np


def verify_labels_are_equal(expected_labels: np.ndarray, actual_labels: np.ndarray, id=""):
    assert expected_labels.dtype == actual_labels.dtype, (
        "make sure reference and target is the same dtype before comparing the labels. \n"
        f"reference is: {expected_labels.dtype} and target is: {actual_labels.dtype}"
    )
    assert np.all(np.isin(actual_labels, expected_labels)), (
        f"Unexpected labels found for {id} \n" f"expected: {expected_labels} \n" f"found: {actual_labels}"
    )


def verify_array_shape_is_equal(reference: np.ndarray, target: np.ndarray, id=""):
    assert np.all(reference.shape == target.shape), (
        f"Sizes do not match for {id}" f"Image is: {reference.shape} while the label is {target.shape}"
    )


def verify_shape_is_equal(reference, target, id=""):
    assert np.all(reference == target), (
        f"Sizes do not match for {id}" f"Image is: {reference.shape} while the label is {target.shape}"
    )



import numpy as np
import logging
import math
import nibabel as nib
import os
import torch
import torch.nn.functional as F
import cc3d
from typing import Union, List, Optional
from yucca.functional.array_operations.normalization import normalizer
from skimage.transform import resize
from yucca.functional.array_operations.cropping_and_padding import pad_to_size, get_pad_kwargs
from yucca.functional.testing.data.nifti import verify_nifti_header_is_valid, verify_orientation_is_LR_PA_IS
from yucca.functional.testing.data.array import verify_shape_is_equal
from yucca.functional.array_operations.transpose import transpose_case, transpose_array
from yucca.functional.array_operations.bounding_boxes import get_bbox_for_foreground
from yucca.functional.array_operations.cropping_and_padding import crop_to_box
from yucca.functional.utils.nib_utils import (
    get_nib_spacing,
    get_nib_orientation,
    reorient_nib_image,
)
from yucca.functional.utils.type_conversions import nifti_or_np_to_np
from yucca.functional.utils.loading import read_file_to_nifti_or_np


def analyze_label(label, enable_connected_components_analysis: bool = False, spacing: list = [], per_class=False):
    # we get some (no need to get all) locations of foreground, that we will later use in the
    # oversampling of foreground classes
    # And we also potentially analyze the connected components of the label
    foreground_locs = get_foreground_locations(label, per_class=per_class, max_locs_total=100000)
    if not enable_connected_components_analysis:
        label_cc_n = 0
        label_cc_sizes = 0
    else:
        numbered_ground_truth, label_cc_n = cc3d.connected_components(label, connectivity=26, return_N=True)
        if len(numbered_ground_truth) == 0:
            label_cc_sizes = 0
        else:
            label_cc_sizes = [i * np.prod(spacing) for i in np.unique(numbered_ground_truth, return_counts=True)[-1][1:]]
    return foreground_locs, label_cc_n, label_cc_sizes


def get_foreground_locations(label, per_class=False, max_locs_total=100000):
    foreground_locations = {}
    if not per_class:
        foreground_locs_for_all = np.array(np.nonzero(label)).T[::10].tolist()
        if len(foreground_locs_for_all) > max_locs_total:
            foreground_locs_for_all = foreground_locs_for_all[:: round(len(foreground_locs_for_all) / max_locs_total)]
        foreground_locations["1"] = foreground_locs_for_all
    else:
        foreground_classes_present = np.unique(label)[1:]
        if len(foreground_classes_present) == 0:
            return foreground_locations
        max_locs_per_class = int(max_locs_total / len(foreground_classes_present))
        for c in foreground_classes_present:
            foreground_locs_for_c = np.array(np.where(label == int(c))).T[::10]
            if len(foreground_locs_for_c) > 0:
                if len(foreground_locs_for_c) > max_locs_per_class:
                    foreground_locs_for_c = foreground_locs_for_c[:: round(len(foreground_locs_for_c) / max_locs_per_class)]
                foreground_locations[str(int(c))] = foreground_locs_for_c
    return foreground_locations


def determine_resample_size_from_target_size(current_size, current_spacing, target_size, keep_aspect_ratio: bool = False):
    if keep_aspect_ratio:
        resample_target_size = np.array(current_size * np.min(target_size / current_size)).astype(int)
        final_target_size = target_size
        final_target_size = [math.ceil(i / 16) * 16 for i in final_target_size]
    else:
        resample_target_size = target_size
        resample_target_size = [math.ceil(i / 16) * 16 for i in resample_target_size]
        final_target_size = None
    new_spacing = (
        (np.array(resample_target_size).astype(float) / current_size.astype(float)) * np.array(current_spacing).astype(float)
    ).tolist()
    return resample_target_size, final_target_size, new_spacing


def determine_resample_size_from_target_spacing(current_size, current_spacing, target_spacing: np.ndarray):
    final_target_size = None
    resample_target_size = np.round((current_spacing / target_spacing).astype(float) * current_size).astype(int)
    new_spacing = target_spacing.tolist()
    return resample_target_size, final_target_size, new_spacing


def resample_and_normalize_case(
    case: list,
    target_size,
    norm_op: str,
    intensities: list = None,
    label: np.ndarray = None,
    allow_missing_modalities: bool = False,
):
    # Normalize and Transpose images to target view.
    # Transpose labels to target view.
    assert len(case) == len(norm_op), (
        "number of images and "
        "normalization  operations does not match. \n"
        f"len(images) == {len(case)} \n"
        f"len(norm_op) == {len(norm_op)} \n"
    )

    for i in range(len(case)):
        image = case[i]
        assert image is not None
        if image.size == 0:
            assert allow_missing_modalities is True, "missing modality and allow_missing_modalities is not enabled"
        else:
            # Normalize
            if intensities is not None:
                case[i] = normalizer(image, scheme=norm_op[i], intensities=intensities[i])
            else:
                case[i] = normalizer(image, scheme=norm_op[i])

            # Resample to target shape and spacing
            try:
                case[i] = resize(case[i], output_shape=target_size, order=3)
            except OverflowError:
                logging.error("Unexpected values in either shape or image for resize")
    if label is not None:
        try:
            label = resize(label, output_shape=target_size, order=0, anti_aliasing=False)
        except OverflowError:
            logging.error("Unexpected values in either shape or label for resize")
        return case, label
    return case


def pad_case_to_size(case: list, size, pad_value="min", label=None):
    for i in range(len(case)):
        pad_kwargs = get_pad_kwargs(data=case[i], pad_value=pad_value)
        case[i], _ = pad_to_size(case[i], size, **pad_kwargs)
    if label is None:
        return case
    label, _ = pad_to_size(label, size, **pad_kwargs)
    return case, label


def apply_nifti_preprocessing_and_return_numpy(
    images,
    original_size,
    target_orientation,
    label=None,
    include_header=False,
    strict=True,
):
    # If qform and sform are both missing the header is corrupt and we do not trust the
    # direction from the affine
    # Make sure you know what you're doing
    metadata = {
        "original_spacing": np.array([1.0] * len(original_size)).tolist(),
        "original_orientation": None,
        "final_direction": None,
        "header": None,
        "affine": None,
        "reoriented": False,
    }

    if isinstance(images[0], nib.Nifti1Image):
        # If qform and sform are both missing the header is corrupt and we do not trust the
        # direction from the affine
        # Make sure you know what you're doing

        if verify_nifti_header_is_valid(images[0]) is True:
            if strict:
                assert verify_orientation_is_LR_PA_IS(
                    images[0]
                ), "unexpected NIFTI axes. Consider RAS-conversion during task conversion"
            metadata["reoriented"] = True
            metadata["original_orientation"] = get_nib_orientation(images[0])
            metadata["final_direction"] = target_orientation
            images = [
                reorient_nib_image(image, metadata["original_orientation"], metadata["final_direction"]) for image in images
            ]
            if label is not None and isinstance(label, nib.Nifti1Image):
                label = reorient_nib_image(label, metadata["original_orientation"], metadata["final_direction"])
        if include_header:
            metadata["header"] = images[0].header
        metadata["original_spacing"] = get_nib_spacing(images[0]).tolist()
        metadata["affine"] = images[0].affine

    images = [nifti_or_np_to_np(image) for image in images]
    if label is not None:
        label = nifti_or_np_to_np(label)
    return images, label, metadata


def determine_target_size(
    images_transposed: list,
    original_spacing,
    transpose_forward,
    target_size,
    target_spacing,
    keep_aspect_ratio,
):
    image_shape_t = np.array(images_transposed[0].shape)
    original_spacing_t = original_spacing[transpose_forward]

    # We do not want to change the aspect ratio so we resample using the minimum alpha required
    # to attain 1 correct dimension, and then the rest will be padded.
    # Additionally we make sure each dimension is divisible by 16 to avoid issues with standard pooling/stride settings
    if target_size is not None:
        resample_target_size, final_target_size, new_spacing = determine_resample_size_from_target_size(
            current_size=image_shape_t,
            current_spacing=original_spacing_t,
            target_size=target_size,
            keep_aspect_ratio=keep_aspect_ratio,
        )

    # Otherwise we need to calculate a new target shape, and we need to factor in that
    # the images will first be transposed and THEN resampled.
    # Find new shape based on the target spacing
    elif target_spacing is not None:
        target_spacing = np.array(target_spacing, dtype=float)
        target_spacing_t = target_spacing[transpose_forward]
        resample_target_size, final_target_size, new_spacing = determine_resample_size_from_target_spacing(
            current_size=image_shape_t, current_spacing=original_spacing_t, target_spacing=target_spacing_t
        )
    else:
        resample_target_size = image_shape_t
        final_target_size = None
        new_spacing = original_spacing_t.tolist()
    return resample_target_size, final_target_size, new_spacing


def preprocess_case_for_training_with_label(
    images: List[Union[np.ndarray, nib.Nifti1Image]],
    label: Union[np.ndarray, nib.Nifti1Image],
    normalization_operation: list,
    allow_missing_modalities: bool = False,
    background_pixel_value: int = 0,
    enable_cc_analysis: bool = False,
    foreground_locs_per_label: bool = False,
    missing_modality_idxs: list = [],
    crop_to_nonzero: bool = True,
    keep_aspect_ratio_when_using_target_size: bool = False,
    image_properties: Optional[dict] = {},
    intensities: Optional[List] = None,
    target_orientation: Optional[str] = "RAS",
    target_size: Optional[List] = None,
    target_spacing: Optional[List] = None,
    transpose: Optional[list] = [0, 1, 2],
):
    """
    one of target_size or target_spacing is required.
    """

    images, label, image_properties["nifti_metadata"] = apply_nifti_preprocessing_and_return_numpy(
        images=images,
        original_size=np.array(images[0].shape),
        target_orientation=target_orientation,
        label=label,
        include_header=False,
    )

    original_size = images[0].shape

    # Cropping is performed to save computational resources. We are only removing background.
    if crop_to_nonzero:
        nonzero_box = get_bbox_for_foreground(images[0], background_label=background_pixel_value)
        image_properties["crop_to_nonzero"] = nonzero_box
        for i in range(len(images)):
            images[i] = crop_to_box(images[i], nonzero_box)
        label = crop_to_box(label, nonzero_box)
    else:
        image_properties["crop_to_nonzero"] = crop_to_nonzero

    image_properties["size_before_transpose"] = list(images[0].shape)

    images = transpose_case(images, axes=transpose)
    label = transpose_array(label, axes=transpose)

    image_properties["size_after_transpose"] = list(images[0].shape)

    resample_target_size, final_target_size, new_spacing = determine_target_size(
        images_transposed=images,
        original_spacing=np.array(image_properties["nifti_metadata"]["original_spacing"]),
        transpose_forward=transpose,
        target_size=target_size,
        target_spacing=target_spacing,
        keep_aspect_ratio=keep_aspect_ratio_when_using_target_size,
    )

    # here we need to make sure missing modalities are accounted for, as the order of the images
    # can matter for normalization operations
    for missing_mod in missing_modality_idxs:
        images.insert(missing_mod, np.array([]))
    images, label = resample_and_normalize_case(
        case=images,
        target_size=resample_target_size,
        norm_op=normalization_operation,
        intensities=intensities,
        label=label,
        allow_missing_modalities=allow_missing_modalities,
    )

    if final_target_size is not None:
        images, label = pad_case_to_size(case=images, size=final_target_size, label=label)
    image_properties["foreground_locations"], image_properties["label_cc_n"], image_properties["label_cc_sizes"] = (
        analyze_label(
            label=label, enable_connected_components_analysis=enable_cc_analysis, per_class=foreground_locs_per_label
        )
    )

    first_existing_modality = list(set(range(len(images))).difference(missing_modality_idxs))[0]
    image_properties["new_size"] = list(images[first_existing_modality].shape)
    image_properties["original_spacing"] = image_properties["nifti_metadata"]["original_spacing"]
    image_properties["original_size"] = original_size
    image_properties["original_orientation"] = image_properties["nifti_metadata"]["original_orientation"]
    image_properties["new_spacing"] = new_spacing
    image_properties["new_direction"] = image_properties["nifti_metadata"]["final_direction"]
    return images, label, image_properties


def preprocess_case_for_training_without_label(
    images: List[Union[np.ndarray, nib.Nifti1Image]],
    normalization_operation: list,
    allow_missing_modalities: bool = False,
    background_pixel_value: int = 0,
    missing_modality_idxs: list = [],
    crop_to_nonzero: bool = True,
    keep_aspect_ratio_when_using_target_size: bool = False,
    image_properties: Optional[dict] = {},
    intensities: Optional[List] = [],
    target_orientation: Optional[str] = "RAS",
    target_size: Optional[List] = None,
    target_spacing: Optional[List] = None,
    transpose: Optional[list] = [0, 1, 2],
):
    """
    one of target_size or target_spacing is required.
    """

    images, label, image_properties["nifti_metadata"] = apply_nifti_preprocessing_and_return_numpy(
        images=images,
        original_size=np.array(images[0].shape),
        target_orientation=target_orientation,
        label=None,
        include_header=False,
    )

    original_size = images[0].shape

    # Cropping is performed to save computational resources. We are only removing background.
    if crop_to_nonzero:
        nonzero_box = get_bbox_for_foreground(images[0], background_label=background_pixel_value)
        image_properties["crop_to_nonzero"] = nonzero_box
        for i in range(len(images)):
            images[i] = crop_to_box(images[i], nonzero_box)
    else:
        image_properties["crop_to_nonzero"] = crop_to_nonzero

    image_properties["size_before_transpose"] = list(images[0].shape)

    images = transpose_case(images, axes=transpose)

    image_properties["size_after_transpose"] = list(images[0].shape)

    resample_target_size, final_target_size, new_spacing = determine_target_size(
        images_transposed=images,
        original_spacing=np.array(image_properties["nifti_metadata"]["original_spacing"]),
        transpose_forward=transpose,
        target_size=target_size,
        target_spacing=target_spacing,
        keep_aspect_ratio=keep_aspect_ratio_when_using_target_size,
    )

    # here we need to make sure missing modalities are accounted for, as the order of the images
    # can matter for normalization operations
    for missing_mod in missing_modality_idxs:
        images.insert(missing_mod, np.array([]))

    images = resample_and_normalize_case(
        case=images,
        target_size=resample_target_size,
        norm_op=normalization_operation,
        intensities=intensities,
        allow_missing_modalities=allow_missing_modalities,
    )

    if final_target_size is not None:
        images = pad_case_to_size(case=images, size=final_target_size, label=None)

    first_existing_modality = list(set(range(len(images))).difference(missing_modality_idxs))[0]
    image_properties["new_size"] = list(images[first_existing_modality].shape)
    image_properties["label_cc_n"] = image_properties["label_cc_sizes"] = 0
    image_properties["foreground_locations"] = []
    image_properties["original_spacing"] = image_properties["nifti_metadata"]["original_spacing"]
    image_properties["original_size"] = original_size
    image_properties["original_orientation"] = image_properties["nifti_metadata"]["original_orientation"]
    image_properties["new_spacing"] = new_spacing
    image_properties["new_direction"] = image_properties["nifti_metadata"]["final_direction"]
    return images, image_properties


def preprocess_case_for_inference(
    crop_to_nonzero,
    keep_aspect_ratio,
    images: list | tuple,
    intensities: list,
    normalization_scheme: list,
    patch_size: tuple,
    target_size,
    target_spacing,
    target_orientation,
    transpose_forward,
    allow_missing_modalities: bool = False,
) -> torch.Tensor:
    assert isinstance(images, (list, tuple)), "image(s) should be a list or tuple, even if only one image is passed"

    image_properties = {}
    ext = images[0][0].split(os.extsep, 1)[1] if isinstance(images[0], tuple) else images[0].split(os.extsep, 1)[1]
    images = [
        read_file_to_nifti_or_np(image[0]) if isinstance(image, tuple) else read_file_to_nifti_or_np(image) for image in images
    ]

    image_properties["image_extension"] = ext
    image_properties["original_shape"] = np.array(images[0].shape)

    assert len(image_properties["original_shape"]) in [
        2,
        3,
    ], "images must be either 2D or 3D for preprocessing"

    images, _, image_properties["nifti_metadata"] = apply_nifti_preprocessing_and_return_numpy(
        images, image_properties["original_shape"], target_orientation=target_orientation, label=None, include_header=True
    )

    image_properties["uncropped_shape"] = np.array(images[0].shape)

    if crop_to_nonzero:
        nonzero_box = get_bbox_for_foreground(images[0], background_label=0)
        for i in range(len(images)):
            images[i] = crop_to_box(images[i], nonzero_box)
        image_properties["nonzero_box"] = nonzero_box

    image_properties["cropped_shape"] = np.array(images[0].shape)

    images = transpose_case(images, axes=transpose_forward)

    resample_target_size, _, _ = determine_target_size(
        images_transposed=images,
        original_spacing=np.array(image_properties["nifti_metadata"]["original_spacing"]),
        transpose_forward=transpose_forward,
        target_size=target_size,
        target_spacing=target_spacing,
        keep_aspect_ratio=keep_aspect_ratio,
    )

    images = resample_and_normalize_case(
        case=images,
        target_size=resample_target_size,
        norm_op=normalization_scheme,
        intensities=intensities,
        label=None,
        allow_missing_modalities=allow_missing_modalities,
    )

    # From this point images are shape (1, c, x, y, z)
    image_properties["resampled_transposed_shape"] = np.array(images[0].shape)

    for i in range(len(images)):
        images[i], padding = pad_to_size(images[i], patch_size)
    image_properties["padded_shape"] = np.array(images[0].shape)
    image_properties["padding"] = padding

    # Stack and fix dimensions
    images = np.stack(images)[np.newaxis]

    return torch.tensor(images, dtype=torch.float32), image_properties


def reverse_preprocessing(crop_to_nonzero, images, image_properties, n_classes, transpose_forward, transpose_backward):
    """
    Expects images to be preprocessed by the "preprocess_case_for_inference" function or similar that
    will produce an image_properties dict with instructions on how to reverse the operations.

    Expected shape of images are:
    (b, c, x, y(, z))

    (1) Initialization: Extract relevant properties from the image_properties dictionary.
    (2) Padding Reversion: Reverse the padding applied during preprocessing.
    (3) Resampling and Transposition Reversion: Resize the images to revert the resampling operation.
    Transpose the images back to the original orientation.
    (4) Cropping Reversion (Optional): If cropping to the nonzero bounding box was applied, revert the cropping operation.
    (5) Return: Return the reverted images as a NumPy array.
    The original orientation of the image will be re-applied when saving the prediction
    """
    image_properties["save_format"] = image_properties.get("image_extension")
    canvas = torch.zeros((1, n_classes, *image_properties["uncropped_shape"]), dtype=images.dtype)
    shape_after_crop = image_properties["cropped_shape"]
    shape_after_crop_transposed = shape_after_crop[transpose_forward]
    pad = image_properties["padding"]

    verify_shape_is_equal(reference=images.shape[2:], target=image_properties["padded_shape"])

    shape = images.shape[2:]
    if len(pad) == 6:
        images = images[
            :,
            :,
            pad[0] : shape[0] - pad[1],
            pad[2] : shape[1] - pad[3],
            pad[4] : shape[2] - pad[5],
        ]
    elif len(pad) == 4:
        images = images[:, :, pad[0] : shape[0] - pad[1], pad[2] : shape[1] - pad[3]]

    verify_shape_is_equal(reference=images.shape[2:], target=image_properties["resampled_transposed_shape"])

    # Here we Interpolate the array to the original size. The shape starts as [H, W (,D)]. For Torch functionality it is changed to [B, C, H, W (,D)].
    # Afterwards it's squeezed back into [H, W (,D)] and transposed to the original direction.
    images = F.interpolate(images, size=shape_after_crop_transposed.tolist(), mode="trilinear").permute(
        [0, 1] + [i + 2 for i in transpose_backward]
    )

    # Now move the tensor to the CPU
    images = images.cpu()

    verify_shape_is_equal(reference=images.shape[2:], target=image_properties["cropped_shape"])

    if crop_to_nonzero:
        bbox = image_properties["nonzero_box"]
        slices = [
            slice(None),
            slice(None),
            slice(bbox[0], bbox[1] + 1),
            slice(bbox[2], bbox[3] + 1),
        ]
        if len(bbox) == 6:
            slices.append(
                slice(bbox[4], bbox[5] + 1),
            )
        canvas[slices] = images
    else:
        canvas = images
    return canvas.numpy(), image_properties



from yucca.pipeline.managers.YuccaManager import YuccaManager
from yucca.data.augmentation.augmentation_presets import no_aug


class YuccaManager_Ghosting(YuccaManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.augmentation_params = no_aug
        self.augmentation_params["motion_ghosting_p_per_sample"] = 1.0



from yucca.data.augmentation.transforms.YuccaTransform import YuccaTransform
from yucca.functional.transforms import gibbs_ringing
import numpy as np


class GibbsRinging(YuccaTransform):
    def __init__(
        self,
        data_key="image",
        p_per_sample: float = 1.0,
        cut_freq=(96, 129),
        axes=(0, 3),
        clip_to_input_range=False,
    ):
        self.data_key = data_key
        self.p_per_sample = p_per_sample
        self.cut_freq = cut_freq
        self.axes = axes
        self.clip_to_input_range = clip_to_input_range

    @staticmethod
    def get_params(cut_freq, axes):
        cut_freq = np.random.randint(*cut_freq)
        axis = np.random.randint(*axes)
        return cut_freq, axis

    def __gibbsRinging__(self, image, num_sample, axis):
        image = gibbs_ringing(image, num_sample=num_sample, axis=axis, clip_to_input_range=self.clip_to_input_range)
        return image

    def __call__(self, packed_data_dict=None, **unpacked_data_dict):
        data_dict = packed_data_dict if packed_data_dict else unpacked_data_dict
        assert (
            len(data_dict[self.data_key].shape) == 5 or len(data_dict[self.data_key].shape) == 4
        ), f"Incorrect data size or shape.\
            \nShould be (b, c, x, y, z) or (b, c, x, y) and is: {data_dict[self.data_key].shape}"

        for b in range(data_dict[self.data_key].shape[0]):
            for c in range(data_dict[self.data_key][b].shape[0]):
                if np.random.uniform() < self.p_per_sample:
                    cut_freq, axis = self.get_params(self.cut_freq, self.axes)
                    data_dict[self.data_key][b, c] = self.__gibbsRinging__(data_dict[self.data_key][b, c], cut_freq, axis)
        return data_dict



import torch
from torch import nn
from yucca.functional.utils.torch_utils import get_available_device
from yucca.networks.utils.get_steps_for_sliding_window import (
    get_steps_for_sliding_window,
)


class YuccaNet(nn.Module):
    def __init__(self):
        super(YuccaNet, self).__init__()
        # Attributes that need to be defined for each architecture
        self.abbreviation: str = None
        self.dimensions: int = None

    def forward(self):
        """
        implement in individual trainers.
        DO NOT INCLUDE FINAL SOFTMAX/SIGMOID ETC.
        WILL BE HANDLED BY LOSS FUNCTIONS
        """
        pass

    def load_state_dict(self, target_state_dict, *args, **kwargs):
        current_state_dict = self.state_dict()
        # filter unnecessary keys
        target_state_dict = {
            k: v
            for k, v in target_state_dict.items()
            if (k in current_state_dict) and (current_state_dict[k].shape == target_state_dict[k].shape)
        }
        super().load_state_dict(target_state_dict, *args, **kwargs)

    def predict(self, mode, data, patch_size, overlap, sliding_window_prediction=True, mirror=False):
        data = data.to(torch.device(get_available_device()))

        if not sliding_window_prediction:
            return self._full_image_predict(data)

        if mode == "3D":
            predict = self._sliding_window_predict3D
        elif mode == "2D":
            predict = self._sliding_window_predict2D

        pred = predict(data, patch_size, overlap)
        if mirror:
            pred += torch.flip(predict(torch.flip(data, (2,)), patch_size, overlap), (2,))
            pred += torch.flip(predict(torch.flip(data, (3,)), patch_size, overlap), (3,))
            pred += torch.flip(predict(torch.flip(data, (2, 3)), patch_size, overlap), (2, 3))
            div = 4
            if mode == "3D":
                pred += torch.flip(predict(torch.flip(data, (4,)), patch_size, overlap), (4,))
                pred += torch.flip(predict(torch.flip(data, (2, 4)), patch_size, overlap), (2, 4))
                pred += torch.flip(predict(torch.flip(data, (3, 4)), patch_size, overlap), (3, 4))
                pred += torch.flip(
                    predict(torch.flip(data, (2, 3, 4)), patch_size, overlap),
                    (2, 3, 4),
                )
                div += 4
            pred /= div
        return pred

    def _full_image_predict(self, data):
        """
        Standard prediction used in cases where models predict on full-size images.
        This is opposed to patch-based predictions where we use a sliding window approach to generate
        full size predictions.
        """
        return self.forward(data)

    def _sliding_window_predict3D(self, data, patch_size, overlap):
        """
        Sliding window prediction implementation
        """
        canvas = torch.zeros(
            (1, self.num_classes, *data.shape[2:]),
            device=torch.device(get_available_device()),
        )

        x_steps, y_steps, z_steps = get_steps_for_sliding_window(data.shape[2:], patch_size, overlap)
        px, py, pz = patch_size

        for xs in x_steps:
            for ys in y_steps:
                for zs in z_steps:
                    # check if out of bounds
                    out = self.forward(data[:, :, xs : xs + px, ys : ys + py, zs : zs + pz])
                    canvas[:, :, xs : xs + px, ys : ys + py, zs : zs + pz] += out
        return canvas

    def _sliding_window_predict2D(self, data, patch_size, overlap):
        """
        Sliding window prediction implementation
        """
        canvas = torch.zeros(
            (1, self.num_classes, *data.shape[2:]),
            device=torch.device(get_available_device()),
        )

        px, py = patch_size

        # If we have 5 dimensions we are working with 3D data, and need to predict each slice.
        if len(data.shape) == 5:
            x_steps, y_steps = get_steps_for_sliding_window(data.shape[3:], patch_size, overlap)
            for idx in range(data.shape[2]):
                for xs in x_steps:
                    for ys in y_steps:
                        out = self.forward(data[:, :, idx, xs : xs + px, ys : ys + py])
                        canvas[:, :, idx, xs : xs + px, ys : ys + py] += out
            return canvas

        # else we proceed with the data as 2D
        x_steps, y_steps = get_steps_for_sliding_window(data.shape[2:], patch_size, overlap)

        for xs in x_steps:
            for ys in y_steps:
                # check if out of bounds
                out = self.forward(data[:, :, xs : xs + px, ys : ys + py])
                canvas[:, :, xs : xs + px, ys : ys + py] += out
        return canvas


class InitWeights_He(object):
    def __init__(self, neg_slope=1e-2):
        self.neg_slope = neg_slope

    def __call__(self, module):
        if (
            isinstance(module, nn.Conv3d)
            or isinstance(module, nn.Conv2d)
            or isinstance(module, nn.ConvTranspose2d)
            or isinstance(module, nn.ConvTranspose3d)
        ):
            module.weight = nn.init.kaiming_normal_(module.weight, a=self.neg_slope)
            if module.bias is not None:
                module.bias = nn.init.constant_(module.bias, 0)



import shutil
from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p, subfolders
from yucca.pipeline.task_conversion.utils import generate_dataset_json
from yucca.paths import yucca_raw_data


def convert(path: str, subdir: str = "WMH"):
    """INPUT DATA - Define input path and suffixes"""
    path = join(path, subdir)

    """ OUTPUT DATA - Define the task name and prefix """
    task_name = "Task006_WMH_Flair"
    task_prefix = "WMH_F"

    datasets = ["Amsterdam", "Singapore", "Utrecht"]
    site = ""

    # Target paths
    target_base = join(yucca_raw_data, task_name)

    target_imagesTr = join(target_base, "imagesTr")
    target_labelsTr = join(target_base, "labelsTr")

    target_imagesTs = join(target_base, "imagesTs")
    target_labelsTs = join(target_base, "labelsTs")

    maybe_mkdir_p(target_imagesTr)
    maybe_mkdir_p(target_labelsTs)
    maybe_mkdir_p(target_imagesTs)
    maybe_mkdir_p(target_labelsTr)

    ###Populate Target Directory###
    for dataset in datasets:
        dataset_path = join(path, dataset)
        if dataset == "Amsterdam":
            tr_folder = "Train_GE3T"
        else:
            tr_folder = "Train"

        train_folder = join(dataset_path, tr_folder)
        test_folder = join(dataset_path, "Test")

        training_samples = subfolders(train_folder, join=False)
        test_samples = subfolders(test_folder, join=False)

        # First we sort the training data
        for sTr in training_samples:
            # Loading relevant modalities and the ground truth
            src_image_file_path = join(train_folder, sTr, "pre", "FLAIR.nii.gz")
            src_label_path = join(train_folder, sTr, "pre", "wmh.nii.gz")
            dst_image_file_path = f"{target_imagesTr}/{task_prefix}_{sTr}_000.nii.gz"
            dst_label_path = f"{target_labelsTr}/{task_prefix}_{sTr}.nii.gz"
            shutil.copy2(src_image_file_path, dst_image_file_path)
            shutil.copy2(src_label_path, dst_label_path)
            # image_file = open(image_file_path, "rb")
            # label = open(label_path, "rb")
            # shutil.copyfileobj(image_file, gzip.open(f"{target_imagesTr}/{task_prefix}_{sTr}_000.nii.gz", "wb"))
            # shutil.copyfileobj(label, gzip.open(f"{target_labelsTr}/{task_prefix}_{sTr}.nii.gz", "wb"))
            # print("------------------------------------------------------------------------------------")
            # print("Reading image at;", src_image_file_path)
            # print("Reading label at;", src_label_path)
            # print()
            # print("Writing image at;", dst_image_file_path)
            # print("Writing label at;", dst_label_path)
            # print("------------------------------------------------------------------------------------")
            # print()

        # Now we sort the test data
        if dataset == "Amsterdam":
            for site in test_samples:
                samples = subfolders(join(test_folder, site), join=False)
                for sTs in samples:
                    src_image_file_path = join(test_folder, site, sTs, "pre", "FLAIR.nii.gz")
                    src_label_path = join(test_folder, site, sTs, "pre", "wmh.nii.gz")
                    dst_image_file_path = f"{target_imagesTs}/{task_prefix}_{sTs}_000.nii.gz"
                    dst_label_path = f"{target_labelsTs}/{task_prefix}_{sTs}.nii.gz"
                    shutil.copy2(src_image_file_path, dst_image_file_path)
                    shutil.copy2(src_label_path, dst_label_path)

                    # image_file = open(join(test_folder, site, sTs, "pre", "FLAIR.nii.gz"), "rb")
                    # label = open(join(test_folder, site, sTs, "pre", "wmh.nii.gz"), "rb")
                    # shutil.copyfileobj(image_file, gzip.open(f"{target_imagesTs}/{task_prefix}_{sTs}_000.nii.gz", "wb"))
                    # shutil.copyfileobj(label, gzip.open(f"{target_labelsTs}/{task_prefix}_{sTs}.nii.gz", "wb"))
        else:
            for sTs in test_samples:
                src_image_file_path = join(test_folder, sTs, "pre", "FLAIR.nii.gz")
                src_label_path = join(test_folder, sTs, "pre", "wmh.nii.gz")
                dst_image_file_path = f"{target_imagesTs}/{task_prefix}_{sTs}_000.nii.gz"
                dst_label_path = f"{target_labelsTs}/{task_prefix}_{sTs}.nii.gz"
                shutil.copy2(src_image_file_path, dst_image_file_path)
                shutil.copy2(src_label_path, dst_label_path)

                # image_file = open(join(test_folder, sTs, "pre", "FLAIR.nii.gz"), "rb")
                # label = open(join(test_folder, sTs, "pre", "wmh.nii.gz"), "rb")
                # shutil.copyfileobj(image_file, gzip.open(f"{target_imagesTs}/{task_prefix}_{sTs}_000.nii.gz", "wb"))
                # shutil.copyfileobj(label, gzip.open(f"{target_labelsTs}/{task_prefix}_{sTs}.nii.gz", "wb"))

    generate_dataset_json(
        join(target_base, "dataset.json"),
        target_imagesTr,
        target_imagesTs,
        ("Flair",),
        labels={0: "background", 1: "WMH", 2: "Other Pathology"},
        dataset_name=task_name,
        license="CC BY-NC 4.0 DEED",
        dataset_description="White Matter Hyperintensity Segmentation Challenge. Flair images only!",
        dataset_reference="https://dataverse.nl/dataset.xhtml?persistentId=doi:10.34894/AECRSD",
    )



import numpy as np
import logging
from yucca.data.augmentation.transforms.YuccaTransform import YuccaTransform
from typing import Union, Iterable
from yucca.functional.transforms import mask_batch


class Masking(YuccaTransform):
    """
    A Yucca transform class for applying masking to input images or image volumes.

    It can be used for masking with a fixed ratio and fixed token size but it can also be used
    to mask a random ratio of each batch and with random token sizes.

    Args:
        mask (bool, optional): Whether to apply masking. Defaults to False.
        data_key (str, optional): Key to access the image data in the input dictionary. Defaults to "image".
        ratio (Union[Iterable[float], float], optional): Ratio of elements to be masked. Can be a single float or a range of floats.
            Defaults to 0.25. If it's a range of floats the masking ratio will be uniformly sampled from the two floats for each batch.
        token_size (Union[Iterable[Union[int, float]], float, int], optional): Size of the mask token. Can be given either as exact
            patch size in the form of a single integer or an iterator of intergers, or it can be given as a ratio of the full volume.
            If token_size is given as a single int (e.g. 4) the mask token size for a 2D image will be [4, 4] and for 3D [4,4,4].
            If token_size is given as a single float (e.g. 0.05) the float will denote the ratio of the token size relative
                to the image volume size. For token size 0.05 with an image of shape [200, 100, 160] token_size will be
                [200*0.05, 100*0.05, 160*0.05] = [10, 5, 8].
            If token_size is given as an iterator of ints (e.g. [16,16,16]) it will remain unchanged.
            If token_size is given as an iterator of two floats they will be uniformly sampled per batch, and then
                otherwise treated as the case with a single float.
    """

    def __init__(
        self,
        mask=False,
        data_key="image",
        ratio: Union[Iterable[float], float] = 0.25,
        token_size: Union[Iterable[Union[float, int]], float, int] = 0.05,
        pixel_value: Union[float, int, str] = 0,
    ):
        self.mask = mask
        self.data_key = data_key
        self.ratio = ratio
        self.token_size = token_size
        self.pixel_value = pixel_value

    @staticmethod
    def get_params(input_shape, data, pixel_value, ratio, token_size):
        if isinstance(pixel_value, str):
            assert pixel_value in ["min", "max"]
            if pixel_value == "min":
                pixel_value = data.min()
            elif pixel_value == "max":
                pixel_value = data.max()
            else:
                print(f"unrecognized pixel value: got {pixel_value}")
        if isinstance(ratio, (tuple, list)):
            # If ratio is a list/tuple it's a range of ratios from which me sample uniformly per batch
            ratio = np.random.uniform(*ratio)
        if isinstance(token_size, float):
            # if token_size is a float it will be treated as the ratio of the input size
            assert 1 > token_size > 0, "if token_size is a float it needs to be between 1 and 0."
            " It will be treated as a ratio to mask for each image dimension."
            if token_size > 0.25:
                logging.warn(
                    "token_size is set to a ratio over 0.25 of the image. " "This is not intended and should be reconsidered."
                )
            token_size = [int(i * token_size) for i in input_shape]
        elif isinstance(token_size, int):
            token_size = [token_size for i in input_shape]
        elif isinstance(token_size, (tuple, list)) and isinstance(token_size[0], float):
            # If we get a
            assert len(token_size) == 2 and np.all(1 > np.array(token_size)) and np.all(np.array(token_size) > 0)
            if np.any(np.array(token_size) > 0.25):
                logging.warn(
                    "token_size is set to a ratio over 0.25 of the image. " "This is not intended and should be reconsidered."
                )
            token_size = [int(i * np.random.uniform(*token_size)) for i in input_shape]
        return pixel_value, ratio, token_size

    def __mask__(self, image_volume, pixel_value, ratio, token_size):
        image_volume = mask_batch(batch=image_volume, pixel_value=pixel_value, ratio=ratio, token_size=token_size)
        return image_volume

    def __call__(self, packed_data_dict=None, **unpacked_data_dict):
        data_dict = packed_data_dict if packed_data_dict else unpacked_data_dict
        assert (
            len(data_dict[self.data_key].shape) == 5 or len(data_dict[self.data_key].shape) == 4
        ), f"Incorrect data size or shape.\
            \nShould be (b, c, x, y, z) or (b, c, x, y) and is: {data_dict[self.data_key].shape}"
        if self.mask:
            pixel_value, ratio, token_size = self.get_params(
                data_dict[self.data_key].shape[2:],
                data_dict[self.data_key],
                self.pixel_value,
                self.ratio,
                self.token_size,
            )
            data_dict[self.data_key] = self.__mask__(data_dict[self.data_key], pixel_value, ratio, token_size)
        return data_dict
